"""
Overview:
    Functions for ansi escaping and unescapes.
    See `ANSI escape code - Wikipedia <https://en.wikipedia.org/wiki/ANSI_escape_code>`_.
"""
import re

__all__ = [
    'ansi_unescape',
]

_ANSI_PATTERN = re.compile(r'\x1B\[\d+(;\d+){0,2}m')


def ansi_unescape(string: str) -> str:
    """
    Overview:
        Unescape ansi string.
        See `ANSI escape code - Wikipedia <https://en.wikipedia.org/wiki/ANSI_escape_code>`_.

    :param string: Original output string.
    :return unescaped: Unescaped ansi string.

    Examples::
        >>> from hbutils.encoding import ansi_unescape
        >>> ansi_unescape("\x1b[1;31mHello")
        'Hello'
        >>> ansi_unescape("\x1b[2;37;41mWorld")
        'World'
    """
    return _ANSI_PATTERN.sub('', string)



import importlib
import os
import pathlib
import subprocess
import sys
from typing import Dict, Optional
from typing import List

import pkg_resources
from packaging.version import Version

__all__ = [
    'package_version',
    'load_req_file', 'pip',
    'check_reqs', 'check_req_file',
    'pip_install', 'pip_install_req_file',
]

PIP_PACKAGES: Dict[str, str] = {}


def _init_pip_packages():
    global PIP_PACKAGES
    PIP_PACKAGES.clear()
    for i in pkg_resources.working_set:
        PIP_PACKAGES[i.key.lower()] = i.version


_init_pip_packages()


def package_version(name: str) -> Optional[Version]:
    """
    Overview:
        Get version of package with given ``name``.

    :param name: Name of the package, case is not sensitive.
    :return: A :class:`packing.version.Version` object. If the package is not installed, return ``None``.

    Examples::
        >>> from hbutils.system import package_version
        >>>
        >>> package_version('pip')
        <Version('21.3.1')>
        >>> package_version('setuptools')
        <Version('59.6.0')>
        >>> package_version('not_a_package')
        None
    """
    _lower_name = name.lower()
    if _lower_name in PIP_PACKAGES:
        return pkg_resources.parse_version(PIP_PACKAGES[_lower_name])
    else:
        return None


def load_req_file(requirements_file: str) -> List[str]:
    """
    Overview:
        Load requirements items from a ``requirements.txt`` file.

    :param requirements_file: Requirements file.
    :return requirements: List of requirements.

    Examples::
        >>> from hbutils.system import load_req_file
        >>> load_req_file('requirements.txt')
        ['packaging>=21.3', 'setuptools>=50.0']
    """
    with pathlib.Path(requirements_file).open() as reqfile:
        return list(map(str, pkg_resources.parse_requirements(reqfile)))


def pip(*args, silent: bool = False):
    """
    Overview:
        Run pip command with code.

    :param args: Command line arguments for ``pip`` command.
    :param silent: Do not print anything. Default is false, which means print the output to ``sys.stdout`` \
        and ``sys.stderr``.

    Examples::
        >>> from hbutils.system import pip
        >>> pip('-V')
        pip 22.3.1 from /home/user/myproject/venv/lib/python3.7/site-packages/pip (python 3.7)
        >>> pip('-V', silent=True)  # nothing will be printed
    """
    try:
        process = subprocess.run(
            [sys.executable, '-m', 'pip', *args],
            stdin=sys.stdin if not silent else None,
            stdout=sys.stdout if not silent else subprocess.PIPE,
            stderr=sys.stderr if not silent else subprocess.PIPE,
        )
        assert not process.returncode, f'Error when calling {process.args!r}{os.linesep}' \
                                       f'Error Code - {process.returncode}{os.linesep}' \
                                       f'Stdout:{os.linesep}' \
                                       f'{process.stdout.decode()}{os.linesep}' \
                                       f'{os.linesep}' \
                                       f'Stderr:{os.linesep}' \
                                       f'{process.stderr.decode()}{os.linesep}'
        process.check_returncode()
    finally:
        if args and args[0] in {'install', 'uninstall'}:
            importlib.reload(pkg_resources)
            _init_pip_packages()


def check_reqs(reqs: List[str]) -> bool:
    """
    Overview:
        Check if the given requirements are all satisfied.

    :param reqs: List of requirements.
    :return satisfied: All the requirements in ``reqs`` satisfied or not.

    Examples::
        >>> from hbutils.system import check_reqs
        >>> check_reqs(['pip>=20.0'])
        True
        >>> check_reqs(['pip~=19.2'])
        False
        >>> check_reqs(['pip>=20.0', 'setuptools>=50.0'])
        True
    """
    try:
        pkg_resources.require(reqs)
    except (pkg_resources.DistributionNotFound, pkg_resources.VersionConflict):
        return False
    else:
        return True


def check_req_file(requirements_file: str) -> bool:
    """
    Overview:
        Check if the requirements in the given ``requirements_file`` is satisfied.

    :param requirements_file: Requirements file, such as ``requirements.txt``.
    :return satisfied: All the requirements in ``requirements_file`` satisfied or not.

    Examples::
        >>> from hbutils.system import check_req_file
        >>>
        >>> check_req_file('requirements.txt')
        True
        >>> check_req_file('requirements-test.txt')
        True
    """
    return check_reqs(load_req_file(requirements_file))


def pip_install(reqs: List[str], silent: bool = False, force: bool = False, user: bool = False):
    """
    Overview:
        Pip install requirements with code.
        Similar to ``pip install req1 req2 ...``.

    :param reqs: Requirement items to install.
    :param silent: Do not print anything. Default is ``False``.
    :param force: Force execute the ``pip install`` command. Default is ``False`` which means the requirements \
        will be checked before installation, and the installation will be only executed when \
        some requirements not installed.
    :param user: User mode, represents ``--user`` option in ``pip``.

    Examples::
        >>> from hbutils.system import pip_install
        >>> pip_install(['scikit-learn'])  # not installed
        Looking in indexes: https://xxx/simple
        Collecting scikit-learn
          Using cached https://xxx/scikit_learn-1.0.2-cp37-cp37m-manylinux_2_17_x86_64.manylinux2014_x86_64.whl (24.8 MB)
        Installing collected packages: threadpoolctl, scipy, joblib, scikit-learn
        Successfully installed joblib-1.2.0 scikit-learn-1.0.2 scipy-1.7.3 threadpoolctl-3.1.0
        >>> pip_install(['numpy>=1.10.0'])  # installed
        >>> pip_install(['numpy>=1.10.0'], force=True)  # force execute
        Looking in indexes: https://xxx/simple
        Requirement already satisfied: numpy>=1.10.0 in ./venv/lib/python3.7/site-packages (1.21.6)
    """
    if force or not check_reqs(reqs):
        pip('install', *(('--user',) if user else ()), *reqs, silent=silent)


def pip_install_req_file(requirements_file: str, silent: bool = False, force: bool = False, user: bool = False):
    """
    Overview:
        Pip install requirements from file with code.
        Similar to ``pip install -r requirements.txt``.

    :param requirements_file: Requirements file, such as ``requirements.txt``.
    :param silent: Do not print anything. Default is ``False``.
    :param force: Force execute the ``pip install`` command. Default is ``False`` which means the requirements \
        will be checked before installation, and the installation will be only executed when \
        some requirements not installed.
    :param user: User mode, represents ``--user`` option in ``pip``.

    Examples::
        >>> from hbutils.system import pip_install_req_file
        >>> pip_install_req_file('requirements.txt')  # pip install -r requirements.txt
    """
    if force or not check_req_file(requirements_file):
        pip('install', *(('--user',) if user else ()), '-r', requirements_file, silent=silent)



"""
Overview:
    Implement of topological sorting algorithm.

    Wikipedia: `Topological sorting <https://en.wikipedia.org/wiki/Topological_sorting>`_.
"""
from heapq import heappush, heappop
from queue import Queue
from typing import TypeVar, Collection, Tuple, List, Iterable, Hashable, Callable, Set

__all__ = ['topoids', 'topo']


def topoids(n: int, edges: Collection[Tuple[int, int]], sort: bool = False) -> List[int]:
    """
    Overview:
        Topological sort with nodes count and edges.

    Arguments:
        - n (:obj:`int`): Count of nodes.
        - edges (:obj:`Collection[Tuple[int, int]]`): Collection of edges, \
            in each tuple (x, y), means x should be earlier appeared than y in the final sequence.
        - sort (:obj:`bool`): Keep the output list in order. When open, the time complexity will \
            increase an extra :math:`O\\left(\\log{N}\\right)` because of the maintenance of heap. \
            Default is ``False``, which means no not keep the strict order.

    Returns:
        - sequence (:obj:`List[int]`): Sorted sequence.

    Example:
        >>> topoids(3, [])
        [0, 1, 2]
        >>> topoids(3, [(0, 1), (2, 1)])
        [0, 2, 1]
        >>> topoids(3, [(0, 1), (2, 1), (1, 0)])
        ArithmeticError: ('Invalid topological graph, for some node ids not accessible - (0, 1).', (0, 1))

        >>> topoids(4, [(0, 2), (0, 1), (2, 3), (1, 3)])
        [0, 1, 2, 3]  # [0, 2, 1, 3] is also possible
        >>> topoids(4, [(0, 2), (0, 1), (2, 3), (1, 3)], sort=True)
        [0, 1, 2, 3]  # only [0, 1, 2, 3] is possible
    """
    if sort:
        queue_init = lambda: []
        queue_push = heappush
        queue_pop = heappop
        queue_empty = lambda q: not q
    else:
        queue_init = lambda: Queue()
        queue_push = lambda q, x: q.put(x)
        queue_pop = lambda q: q.get()
        queue_empty = lambda q: q.empty()

    if n == 0:
        return []

    in_degree: List[int] = [0] * n
    goings: List[Set[int]] = [set() for _ in range(n)]
    for arrow_tail, arrow_head, in list(set(edges)):
        assert isinstance(arrow_tail, int) and 0 <= arrow_tail < n, \
            f'Tail should be in [0, {n}) but {arrow_tail} found.'
        assert isinstance(arrow_head, int) and 0 <= arrow_head < n, \
            f'Head should be in [0, {n}) but {arrow_head} found.'
        in_degree[arrow_head] += 1
        goings[arrow_tail].add(arrow_head)

    queue = queue_init()
    visited = []
    for i in range(n):
        if in_degree[i] == 0:
            queue_push(queue, i)

    while not queue_empty(queue):
        arrow_tail = queue_pop(queue)
        visited.append(arrow_tail)
        for arrow_head in goings[arrow_tail]:
            assert in_degree[arrow_head] > 0
            in_degree[arrow_head] -= 1
            if in_degree[arrow_head] == 0:
                queue_push(queue, arrow_head)

    if len(visited) < n:
        missing = tuple(sorted(set(range(n)) - set(visited)))
        raise ArithmeticError(f'Invalid topological graph, '
                              f'for some node ids not accessible - {repr(missing)}.', missing)

    return visited


_ElementType = TypeVar('_ElementType')
_IdType = TypeVar('_IdType', bound=Hashable)


# noinspection PyUnresolvedReferences
def topo(items: Iterable[_ElementType],
         edges: Collection[Tuple[_ElementType, _ElementType]],
         identifier: Callable[[_ElementType], _IdType] = None,
         sort: bool = False) -> List[_ElementType]:
    """
    Overview:
        Topological sort with objects and their edges.

    Arguments:
        - items (:obj:`Iterable[_ElementType]`): List of the items.
        - edges (:obj:`Collection[Tuple[_ElementType, _ElementType]]`): Collection of edges, \
            in each tuple (x, y), means x should be earlier appeared than y in the final sequence.
        - identifier (:obj:`Callable[[_ElementType], _IdType]`): Identifier function for the items, \
            need to make sure the output id value is hashable. Default is ``None``, which means \
            use the id value of the objects.
        - sort (:obj:`bool`): Keep the output list in order. When open, the time complexity will \
            increase an extra :math:`O\\left(\\log{N}\\right)` because of the maintenance of heap. \
            Default is ``False``, which means no not keep the strict order.

    Returns:
        - sequence (:obj:`List[_ElementType]`): Sorted sequence.

    Examples::

        >>> n1 = _Container(1)  # _Container is a hashable wrapper class
        >>> n2 = _Container('sdfklj')
        >>> n3 = _Container((2, 3))
        >>> n4 = _Container((3, 'sdj'))
        >>> n5 = _Container(1)
        >>> topo([n1, n2, n3], [], sort=True)
        [n1, n2, n3]
        >>> topo([n1, n2, n5], [(n1, n2), (n5, n2)], sort=True)
        [n1, n5, n2]
        >>> topo([n1, n2, n5], [(n1, n2), (n5, n2)], identifier=lambda x: x.v, sort=True)
        [n1, n2]
        >>> topo([n1, n2, n3, n4], [(n1, n3), (n3, n1), (n2, n3), (n4, n1)])
        ArithmeticError: ('Invalid topological graph, for some items not accessible - (n1, n3).', (n1, n3))
    """
    identifier = identifier or id
    items = list(items)
    id_map, item_map, n = {}, [], 0
    for index, item in enumerate(items):
        idf = identifier(item)
        if idf not in id_map.keys():
            id_map[idf] = n
            item_map.append(item)
            n += 1

    def item_to_id(it):
        return id_map[identifier(it)]

    edges = [
        (item_to_id(tail), item_to_id(head))
        for tail, head in edges
    ]

    try:
        visited_ids = topoids(n, edges, sort=sort)
    except ArithmeticError as err:
        _, missing_ids = err.args
        missing_items = tuple([item_map[i] for i in sorted(missing_ids)])
        raise ArithmeticError(f'Invalid topological graph, '
                              f'for some items not accessible - {repr(missing_items)}.', missing_items)
    else:
        return [item_map[i] for i in visited_ids]



"""
Overview:
    Useful utilities for pluralize your words.
"""

from .inflection import pluralize, singularize

__all__ = [
    'plural_form', 'plural_word', 'singular_form'
]


def plural_form(word: str) -> str:
    """
    Overview:
        Get the pluralized form of the given ``word``.

        The same as :func:`hbutils.string.inflection.pluralize`.

    Arguments:
        - word (:obj:`str`): The given word to be pluralized.

    Returns:
        - pluralized (:obj:`str`): Pluralized word.

    Examples::
        >>> from hbutils.string import plural_form
        >>> plural_form('it')
        'they'
        >>> plural_form('word')
        'words'
        >>> plural_form('woman')
        'women'
    """
    return pluralize(word)


def singular_form(word: str) -> str:
    """
    Overview:
        Get the singular form of the given ``word``.

        The same as :func:`hbutils.string.inflection.singularize`.

    Arguments:
        - word (:obj:`str`): The given word to be singularized.

    Returns:
        - single (:obj:`str`): Singular form of word.

    Examples::
        >>> from hbutils.string import singular_form
        >>> singular_form('they')
        'it'
        >>> singular_form('it')
        'it'
        >>> singular_form('women')
        'woman'
        >>> singular_form('words')
        'word'
        >>> singular_form('themselves')
        'itself'
    """
    return singularize(word)


def plural_word(count: int, word: str) -> str:
    """
    Overview:
        Get plural form of the whole word, with the number before the word.

    Arguments:
        - count (:obj:`int`): Count of the word, should be a non-negative integer.
        - word (:obj:`str`): Word to be pluralized.

    Returns:
        - plural_word (:obj:`str`): Pluralized word, with the number.

    Examples::
        >>> from hbutils.string import plural_word
        >>> plural_word(0, 'word')
        '0 words'
        >>> plural_word(1, 'word')
        '1 word'
        >>> plural_word(2, 'word')
        '2 words'
        >>> plural_word(20, 'word')
        '20 words'
        >>> plural_word(233, 'word')
        '233 words'
    """
    single_word = singular_form(word)
    if count != 1:
        return f'{count} {plural_form(single_word)}'
    else:
        return f'{count} {single_word}'



"""
Overview:
    Capture for outputs.
"""
import io
import os
import pathlib
from contextlib import redirect_stdout, redirect_stderr, contextmanager
from threading import Lock
from typing import ContextManager, Optional

__all__ = [
    'OutputCaptureResult',
    'capture_output', 'disable_output',
]

from .._base import TemporaryDirectory


class OutputCaptureResult:
    """
    Overview:
        Result model of output capturing.
    """

    def __init__(self):
        """
        Constructor of :class:`OutputCaptureResult`.
        """
        self._stdout = None
        self._stderr = None
        self._lock = Lock()
        self._lock.acquire()

    def put_result(self, stdout: Optional[str], stderr: Optional[str]):
        """
        Put result inside this model.

        :param stdout: Stdout result.
        :param stderr: Stderr result.
        """
        self._stdout, self._stderr = stdout, stderr
        self._lock.release()

    @property
    def stdout(self) -> Optional[str]:
        """
        Stdout of the output result.

        .. note::
            Do not use this property when :func:`capture_output`'s with block is not quited, \
            or this property will be jammed due to the deadlock inside.
        """
        with self._lock:
            return self._stdout

    @property
    def stderr(self) -> Optional[str]:
        """
        Stderr of the output result.

        .. note::
            Do not use this property when :func:`capture_output`'s with block is not quited, \
            or this property will be jammed due to the deadlock inside.
        """
        with self._lock:
            return self._stderr


@contextmanager
def _capture_via_memory() -> ContextManager[OutputCaptureResult]:
    r = OutputCaptureResult()
    with io.StringIO() as sout, io.StringIO() as serr:
        try:
            with redirect_stdout(sout), redirect_stderr(serr):
                yield r
        finally:
            r.put_result(
                sout.getvalue(),
                serr.getvalue(),
            )


@contextmanager
def _capture_via_tempfile() -> ContextManager[OutputCaptureResult]:
    r = OutputCaptureResult()
    with TemporaryDirectory() as tdir:
        stdout_file = os.path.join(tdir, 'stdout')
        stderr_file = os.path.join(tdir, 'stderr')
        try:
            with open(stdout_file, 'w+') as f_stdout, open(stderr_file, 'w+') as f_stderr:
                with redirect_stdout(f_stdout), redirect_stderr(f_stderr):
                    yield r
        finally:
            r.put_result(
                pathlib.Path(stdout_file).read_text(),
                pathlib.Path(stderr_file).read_text(),
            )


@contextmanager
def capture_output(mem: bool = False) -> ContextManager[OutputCaptureResult]:
    """
    Overview:
        Capture all the output to ``sys.stdout`` and ``sys.stderr`` in this ``with`` block.

    :param mem: Use memory to put the result or not. Default is ``False`` \
        which means the output will be redirected to temporary files.

    Examples::
        >>> from hbutils.testing import capture_output
        >>> import sys
        >>>
        >>> with capture_output() as r:
        ...     print('this is stdout')
        ...     print('this is stderr', file=sys.stderr)
        ...
        >>> r.stdout
        'this is stdout\\n'
        >>> r.stderr
        'this is stderr\\n'

    .. note::
        When ``mem`` is set to ``True``, :class:`io.StringIO` is used, which do not have ``fileno`` \
            method. This may cause some problems in some cases (such as :func:`subprocess.run`).

    """
    mock_func = _capture_via_memory if mem else _capture_via_tempfile
    with mock_func() as co:
        yield co


@contextmanager
def disable_output() -> ContextManager[OutputCaptureResult]:
    """
    Overview:
        Disable all the output to ``sys.stdout`` and ``sys.stderr`` in this ``with`` block.

    Examples::
        >>> import sys
        >>> from hbutils.testing import disable_output
        >>>
        >>> with disable_output():  # no output will be shown
        ...     print('this is stdout')
        ...     print('this is stderr', file=sys.stderr)
    """
    with open(os.devnull, 'w') as sout:
        with redirect_stdout(sout), redirect_stderr(sout):
            yield



import io
from typing import BinaryIO, Optional

from .base import CIOType, CFixedType
from .buffer import c_buffer
from ..encoding import auto_decode

__all__ = [
    'CStringType',
    'c_str',
    'CSizedStringType',
    'c_sized_str',
]


def _auto_encode(s: str, encoding) -> bytes:
    return s.encode(encoding if encoding is not None else 'utf-8')


class CStringType(CIOType):
    """
    Overview:
        Simple string type.

        It should end with a single ``\\x00``, which is quite common in C language.
    """

    def __init__(self, encoding=None):
        """
        Constructor of :class:`CStringType`.

        :param encoding: Encoding type, default is ``None`` which means auto-detect the encodings.
        """
        self.__encoding = encoding

    @property
    def encoding(self) -> Optional[str]:
        """
        Encoding type.
        """
        return self.__encoding

    def read(self, file: BinaryIO) -> str:
        """
        Read simple string value.

        :param file: Binary file, ``io.BytesIO`` is supported as well.
        :return: String value.
        """
        b = bytearray()
        while True:
            bt = file.read(1)[0]
            if bt:
                b.append(bt)
            else:
                break

        return auto_decode(bytes(b), self.__encoding)

    def write(self, file: BinaryIO, val: str):
        """
        Write simple string value to binary IO object.

        :param file: Binary file, ``io.BytesIO`` is supported as well.
        :param val: String to write.
        """
        if not isinstance(val, str):
            raise TypeError(f'String value expected, but {repr(val)} found.')

        file.write(_auto_encode(val, self.__encoding) + b'\x00')


c_str = CStringType()
"""
Overview:
    Reading and writing simple string, ends with a single ``\\x00``.

Examples::
    >>> import io
    >>> from hbutils.binary import c_str
    >>> 
    >>> with io.BytesIO(
    ...         b'kdsfjldsjflkdsmgds\\x00'
    ...         b'\\xd0\\x94\\xd0\\xbe\\xd0\\xb1\\xd1\\x80\\xd1\\x8b\\xd0\\xb9 \\xd0'
    ...         b'\\xb2\\xd0\\xb5\\xd1\\x87\\xd0\\xb5\\xd1\\x80\\x00'
    ...         b'\\xa4\\xb3\\xa4\\xf3\\xa4\\xd0\\xa4\\xf3\\xa4\\xcf\\x00'
    ...         b'\\xcd\\xed\\xc9\\xcf\\xba\\xc3\\x00'
    ... ) as file:
    ...     print(c_str.read(file))
    ...     print(c_str.read(file))
    ...     print(c_str.read(file))
    ...     print(c_str.read(file))
    kdsfjldsjflkdsmgds
    Добрый вечер
    こんばんは
    晚上好
    >>> with io.BytesIO() as file:
    ...     c_str.write(file, "kdsfjld")
    ...     c_str.write(file, "Добрый")
    ...     print(file.getvalue())
    b'kdsfjld\\x00\\xd0\\x94\\xd0\\xbe\\xd0\\xb1\\xd1\\x80\\xd1\\x8b\\xd0\\xb9\\x00'
"""


class CSizedStringType(CFixedType):
    """
    Overview:
        Sized string type.

        It should have a fixed size, which is defined like ``char s[size]`` in C language.
    """

    def __init__(self, size: int, encoding=None):
        """
        Constructor of :class:`CStringType`.

        :param size: Size of the string's space.
        :param encoding: Encoding type, default is ``None`` which means auto-detect the encodings.
        """
        CFixedType.__init__(self, size)
        self.__encoding = encoding
        self._buffer = c_buffer(size)

    @property
    def encoding(self) -> Optional[str]:
        """
        Encoding type.
        """
        return self.__encoding

    def read(self, file: BinaryIO) -> str:
        """
        Read sized string value.

        :param file: Binary file, ``io.BytesIO`` is supported as well.
        :return: String value.
        """
        bytes_ = self._buffer.read(file)
        with io.BytesIO(bytes_ + b'\x00') as bf:
            return c_str.read(bf)

    def write(self, file: BinaryIO, val: str):
        """
        Write sized string value to binary IO object.

        :param file: Binary file, ``io.BytesIO`` is supported as well.
        :param val: String to write.
        """
        if not isinstance(val, str):
            raise TypeError(f'String value expected, but {repr(val)} found.')

        self._buffer.write(file, _auto_encode(val, self.__encoding))


def c_sized_str(size: int) -> CSizedStringType:
    """
    Overview:
        Reading and writing sized string, which occupy a fixed space..

    :param size: Size of the string's space.

    Examples::
        >>> import io
        >>> from hbutils.binary import c_sized_str
        >>> 
        >>> with io.BytesIO(
        ...         b'kdsfjld\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00'
        ...         b'\\xd0\\x94\\xd0\\xbe\\xd0\\xb1\\xd1\\x80\\xd1\\x8b\\xd0\\xb9\\x00\\x00\\x00'
        ... ) as file:
        ...     print(c_sized_str(15).read(file))
        ...     print(c_sized_str(15).read(file))
        kdsfjld
        Добрый
        >>> with io.BytesIO() as file:
        ...     c_sized_str(15).write(file, "kdsfjld")
        ...     c_sized_str(15).write(file, "Добрый")
        ...     print(file.getvalue())
        b'kdsfjld\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\xd0\\x94\\xd0\\xbe\\xd0\\xb1\\xd1\\x80\\xd1\\x8b\\xd0\\xb9\\x00\\x00\\x00'
    """
    return CSizedStringType(size)



"""
Overview:
    Useful utilities for template a string.
"""
from string import Template
from typing import Optional, Mapping, Any

from ..design import SingletonMark

__all__ = [
    'env_template'
]

_NO_DEFAULT_VALUE = SingletonMark('_NO_DEFAULT_VALUE')


def env_template(template: str, environ: Optional[Mapping[str, Any]] = None,
                 safe: bool = False, default: Any = _NO_DEFAULT_VALUE) -> str:
    """
    Overview:
        Mapping all the environment values (not system environment variables) into a template string.

    Arguments:
        - template (:obj:`str`): Template string.
        - environ (:obj:`Optional[Mapping[str, str]]`): Environment values, should be a mapping.
        - safe (:obj:`bool`): Safe substitute, default is ``False`` which means \
            all the value used in template should be able to found in the given ``environ``.
        - default (:obj:`Any`): Default value when no variable provided. Default is ``_NO_DEFAULT_VALUE``, which \
            means ``KeyError`` will be raised.

    Returns:
        - result (:obj:`str`): Substituted string.

    Examples::
        >>> from hbutils.string import env_template
        >>> env_template('${A} + 1 = ${B}', {'A': '1', 'B': '2'})
        '1 + 1 = 2'
        >>> env_template('${A} + 1 = ${B}', {'A': '1'})
        KeyError: 'B'
        >>> env_template('${A} + 1 = ${B}', {'A': '1'}, safe=True)
        '1 + 1 = ${B}'
        >>> env_template('${A} + 1 = ${B}', {'A': '1'}, default='')  # like environment variable
        '1 + 1 = '
    """

    class _DefaultDict(dict):
        def __getitem__(self, item):
            return dict.get(self, item, default)

    _template = Template(template)
    env = environ or {}
    if default is not _NO_DEFAULT_VALUE:
        env = _DefaultDict(env)

    _func = _template.safe_substitute if safe else _template.substitute
    # noinspection PyArgumentList
    return _func(env)



"""
Overview:
    Color model, include rgb, hsv, hls color system.
    
    More color system will be supported soon.
"""
import colorsys
import math
import re
from typing import Optional, Union, Tuple

from .base import _name_to_hex, _CSS3_NAME_MAPS
from ..reflection.func import post_process, raising, freduce, dynamic_call, warning_

__all__ = ['Color']


def _round_mapper(min_: float, max_: float):
    min_, max_ = min(min_, max_), max(min_, max_)
    round_ = max_ - min_

    def _func(v):
        if v < min_:
            v += math.ceil((min_ - v) / round_) * round_
        if v > max_:
            v -= math.ceil((v - max_) / round_) * round_

        return v

    return _func


def _range_mapper(min_: Optional[float], max_: Optional[float], warning=None):
    if min_ is not None and max_ is not None:
        min_, max_ = min(min_, max_), max(min_, max_)
    warning = dynamic_call(warning_(warning if warning is not None else lambda: None))

    def _func(v):
        if max_ is not None and v > max_:
            warning(v, min_, max_)
            return max_
        elif min_ is not None and v < min_:
            warning(v, min_, max_)
            return min_
        else:
            return v

    return _func


class GetSetProxy:
    def __init__(self, getter, setter=None):
        self.__getter = getter
        self.__setter = setter or raising(lambda x: NotImplementedError)

    def set(self, value):
        return self.__setter(value)

    def get(self):
        return self.__getter()


_r_mapper = _range_mapper(0.0, 1.0, lambda v, min_, max_: Warning(
    'Red value should be no less than %.3d and no more than %.3d, but %.3d found.' % (min_, max_, v)))
_g_mapper = _range_mapper(0.0, 1.0, lambda v, min_, max_: Warning(
    'Green value should be no less than %.3d and no more than %.3d, but %.3d found.' % (min_, max_, v)))
_b_mapper = _range_mapper(0.0, 1.0, lambda v, min_, max_: Warning(
    'Blue value should be no less than %.3d and no more than %.3d, but %.3d found.' % (min_, max_, v)))
_a_mapper = _range_mapper(0.0, 1.0, lambda v, min_, max_: Warning(
    'Alpha value should be no less than %.3d and no more than %.3d, but %.3d found.' % (min_, max_, v)))


class RGBColorProxy:
    """
    Overview:
        Color proxy for RGB space.
    """

    def __init__(self, this: 'Color', r: GetSetProxy, g: GetSetProxy, b: GetSetProxy):
        """
        Constructor of :class:`RGBColorProxy`.

        :param this: Original color object.
        :param r: Get-set proxy for red.
        :param g: Get-set proxy for green.
        :param b: Get-set proxy for blue.
        """
        self.__this = this
        self.__rp = r
        self.__gp = g
        self.__bp = b

    @property
    def red(self) -> float:
        """
        Red value (within :math:`\\left[0.0, 1.0\\right]`).

        .. note::
            Setter is available, the change will affect the :class:`Color` object.
        """
        return self.__rp.get()

    @red.setter
    def red(self, new: float):
        self.__rp.set(new)

    @property
    def green(self) -> float:
        """
        Green value (within :math:`\\left[0.0, 1.0\\right]`).

        .. note::
            Setter is available, the change will affect the :class:`Color` object.
        """
        return self.__gp.get()

    @green.setter
    def green(self, new: float):
        self.__gp.set(new)

    @property
    def blue(self) -> float:
        """
        Blue value (within :math:`\\left[0.0, 1.0\\right]`).

        .. note::
            Setter is available, the change will affect the :class:`Color` object.
        """
        return self.__bp.get()

    @blue.setter
    def blue(self, new: float):
        self.__bp.set(new)

    def __iter__(self):
        """
        Iterator for this proxy.

        Examples::
            >>> from hbutils.color import Color
            >>>
            >>> c = Color('green')
            >>> r, g, b = c.rgb
            >>> print(r, g, b)
            0.0 0.5019607843137255 0.0
        """
        yield self.red
        yield self.green
        yield self.blue

    def __repr__(self):
        """
        Representation format.

        Examples::
            >>> from hbutils.color import Color
            >>>
            >>> c = Color('green')
            >>> c.rgb
            <RGBColorProxy red: 0.000, green: 0.502, blue: 0.000>
        """
        return '<{cls} red: {red}, green: {green}, blue: {blue}>'.format(
            cls=self.__class__.__name__,
            red='%.3f' % (self.red,),
            green='%.3f' % (self.green,),
            blue='%.3f' % (self.blue,),
        )


_hsv_h_mapper = _round_mapper(0.0, 1.0)
_hsv_s_mapper = _range_mapper(0.0, 1.0, lambda v, min_, max_: Warning(
    'Saturation value should be no less than %.3d and no more than %.3d, but %.3d found.' % (min_, max_, v)))
_hsv_v_mapper = _range_mapper(0.0, 1.0, lambda v, min_, max_: Warning(
    'Brightness(value) value should be no less than %.3d and no more than %.3d, but %.3d found.' % (min_, max_, v)))


class HSVColorProxy:
    """
    Overview:
        Color proxy for HSV space.
    """

    def __init__(self, this: 'Color', h: GetSetProxy, s: GetSetProxy, v: GetSetProxy):
        """
        Constructor of :class:`HSVColorProxy`.

        :param this: Original color object.
        :param h: Get-set proxy for hue.
        :param s: Get-set proxy for saturation.
        :param v: Get-set proxy for value.
        """
        this.__this = this
        self.__hp = h
        self.__sp = s
        self.__vp = v

    @property
    def hue(self) -> float:
        """
        Hue value (within :math:`\\left[0.0, 1.0\\right]`).

        .. note::
            Setter is available, the change will affect the :class:`Color` object.
        """
        return self.__hp.get()

    @hue.setter
    def hue(self, new: float):
        self.__hp.set(new)

    @property
    def saturation(self) -> float:
        """
        Saturation value (within :math:`\\left[0.0, 1.0\\right]`).

        .. note::
            Setter is available, the change will affect the :class:`Color` object.
        """
        return self.__sp.get()

    @saturation.setter
    def saturation(self, new: float):
        self.__sp.set(new)

    @property
    def value(self) -> float:
        """
        Value value (within :math:`\\left[0.0, 1.0\\right]`).

        .. note::
            Setter is available, the change will affect the :class:`Color` object.
        """
        return self.__vp.get()

    @value.setter
    def value(self, new: float):
        self.__vp.set(new)

    @property
    def brightness(self) -> float:
        """
        Alias for ``value``.
        """
        return self.value

    @brightness.setter
    def brightness(self, new: float):
        self.value = new

    def __iter__(self):
        """
        Iterator for this proxy.

        Examples::
            >>> from hbutils.color import Color
            >>>
            >>> c = Color('green')
            >>> h, s, v = c.hsv
            >>> print(h, s, v)
            0.3333333333333333 1.0 0.5019607843137255
        """
        yield self.hue
        yield self.saturation
        yield self.value

    def __repr__(self):
        """
        Representation format.

        Examples::
            >>> from hbutils.color import Color
            >>>
            >>> c = Color('green')
            >>> c.hsv
            <HSVColorProxy hue: 0.333, saturation: 1.000, value: 0.502>
        """
        return '<{cls} hue: {hue}, saturation: {saturation}, value: {value}>'.format(
            cls=self.__class__.__name__,
            hue='%.3f' % (self.hue,),
            saturation='%.3f' % (self.saturation,),
            value='%.3f' % (self.value,),
        )


_hls_h_mapper = _round_mapper(0.0, 1.0)
_hls_l_mapper = _range_mapper(0.0, 1.0, lambda v, min_, max_: Warning(
    'Lightness value should be no less than %.3d and no more than %.3d, but %.3d found.' % (min_, max_, v)))
_hls_s_mapper = _range_mapper(0.0, 1.0, lambda v, min_, max_: Warning(
    'Saturation value should be no less than %.3d and no more than %.3d, but %.3d found.' % (min_, max_, v)))


class HLSColorProxy:
    """
    Overview:
        Color proxy for HLS space.
    """

    def __init__(self, this: 'Color', h: GetSetProxy, l: GetSetProxy, s: GetSetProxy):
        """
        Constructor of :class:`HLSColorProxy`.

        :param this: Original color object.
        :param h: Get-set proxy for hue.
        :param l: Get-set proxy for lightness.
        :param s: Get-set proxy for saturation.
        """
        this.__this = this
        self.__hp = h
        self.__lp = l
        self.__sp = s

    @property
    def hue(self) -> float:
        """
        Hue value (within :math:`\\left[0.0, 1.0\\right]`).

        .. note::
            Setter is available, the change will affect the :class:`Color` object.
        """
        return self.__hp.get()

    @hue.setter
    def hue(self, new: float):
        self.__hp.set(new)

    @property
    def lightness(self) -> float:
        """
        Lightness value (within :math:`\\left[0.0, 1.0\\right]`).

        .. note::
            Setter is available, the change will affect the :class:`Color` object.
        """
        return self.__lp.get()

    @lightness.setter
    def lightness(self, new: float):
        self.__lp.set(new)

    @property
    def saturation(self) -> float:
        """
        Saturation value (within :math:`\\left[0.0, 1.0\\right]`).

        .. note::
            Setter is available, the change will affect the :class:`Color` object.
        """
        return self.__sp.get()

    @saturation.setter
    def saturation(self, new: float):
        self.__sp.set(new)

    def __iter__(self):
        """
        Iterator for this proxy.

        Examples::
            >>> from hbutils.color import Color
            >>>
            >>> c = Color('green')
            >>> h, l, s = c.hls
            >>> print(h, l, s)
            0.3333333333333333 0.25098039215686274 1.0
        """
        yield self.hue
        yield self.lightness
        yield self.saturation

    def __repr__(self):
        """
        Representation format.

        Examples::
            >>> from hbutils.color import Color
            >>>
            >>> c = Color('green')
            >>> c.hls
            <HLSColorProxy hue: 0.333, lightness: 0.251, saturation: 1.000>
        """
        return '<{cls} hue: {hue}, lightness: {lightness}, saturation: {saturation}>'.format(
            cls=self.__class__.__name__,
            hue='%.3f' % (self.hue,),
            lightness='%.3f' % (self.lightness,),
            saturation='%.3f' % (self.saturation,),
        )


_ratio_to_255 = lambda x: int(round(x * 255))
_ratio_to_hex = post_process(lambda x: '%02x' % (x,))(_ratio_to_255)
_hex_to_255 = lambda x: int(x, base=16) if x is not None else None
_hex_to_ratio = post_process(lambda x: x / 255.0 if x is not None else None)(_hex_to_255)

_RGB_COLOR_PATTERN = re.compile(r'^#?([a-fA-F\d]{2})([a-fA-F\d]{2})([a-fA-F\d]{2})([a-fA-F\d]{2}|)$')


@freduce(init=None)
def _ratio_or(a, b):
    return b if a is None else a


class Color:
    """
    Overview:
        Color utility object.

    Examples::
        >>> from hbutils.color import Color
        >>>
        >>> c = Color('red')  # from name
        >>> c
        <Color red>
        >>> str(c)  # hex format
        '#ff0000'
        >>> (c.rgb.red, c.rgb.green, c.rgb.blue)            # rgb format
        (1.0, 0.0, 0.0)
        >>> (c.hls.hue, c.hls.lightness, c.hls.saturation)  # hls format
        (0.0, 0.5, 1.0)
        >>> (c.hsv.hue, c.hsv.value, c.hsv.saturation)      # hsv format
        (0.0, 1.0, 1.0)


        >>> c1 = Color('#56a3f0')  # from hex
        >>> c1
        <Color #56a3f0>
        >>> str(c1)  # hex format
        '#56a3f0'
        >>> (c1.rgb.red, c1.rgb.green, c1.rgb.blue)            # rgb format
        (0.33725490196078434, 0.6392156862745098, 0.9411764705882353)
        >>> (c1.hls.hue, c1.hls.lightness, c1.hls.saturation)  # hls format
        (0.5833333333333334, 0.6392156862745098, 0.8369565217391304)
        >>> (c1.hsv.hue, c1.hsv.value, c1.hsv.saturation)      # hsv format
        (0.5833333333333334, 0.9411764705882353, 0.6416666666666666)


        >>> c2 = Color('#56a3f077')  # from hex
        >>> c2
        <Color #56a3f0, alpha: 0.467>
        >>> c2.alpha  # alpha value
        0.4666666666666667
        >>> str(c2)   # hex format
        '#56a3f077'
    """

    def __init__(self, c: Union[str, Tuple[float, float, float], 'Color'], alpha: Optional[float] = None):
        """
        Overview:
            Constructor of ``Color``.

        Arguments:
            - c (:obj:`Union[str, Tuple[float, float, float]]`): Color value, can be hex string value \
                or tuple rgb value.
            - alpha: (:obj:`Optional[float]`): Alpha value of color, \
                default is `None` which means no alpha value.
        """
        if isinstance(c, tuple):
            self.__r, self.__g, self.__b = _r_mapper(c[0]), _g_mapper(c[1]), _b_mapper(c[2])
            self.__alpha = _a_mapper(alpha) if alpha is not None else None
        elif isinstance(c, Color):
            self.__init__(str(c), alpha)
        elif isinstance(c, str):
            if _RGB_COLOR_PATTERN.fullmatch(c):
                _rgb_hex = c
            else:
                try:
                    _rgb_hex = _name_to_hex(c)
                except ValueError:
                    raise ValueError("Invalid string color, matching of pattern {pattern} or english name "
                                     "expected but {actual} found.".format(pattern=repr(_RGB_COLOR_PATTERN.pattern),
                                                                           actual=repr(c), ))

            _finding = _RGB_COLOR_PATTERN.findall(_rgb_hex)
            _first = _finding[0]
            rs, gs, bs, as_ = _first
            as_ = None if not as_ else as_

            r, g, b, a = map(_hex_to_ratio, (rs, gs, bs, as_))
            if alpha is not None:
                a = alpha

            self.__init__((r, g, b), a)

        else:
            raise TypeError('Unknown color value - {c}.'.format(c=repr(c)))

    @property
    def alpha(self) -> Optional[float]:
        """
        Alpha value, which means the transparent ratio (within :math:`\\left[0.0, 1.0\\right]`).

        :: note::
            Setter is available.
        """
        return self.__alpha

    @alpha.setter
    def alpha(self, new: Optional[float]):
        if new is not None:
            new = _a_mapper(new)
        self.__alpha = new

    def __get_rgb(self):
        return self.__r, self.__g, self.__b

    def __set_rgb(self, r=None, g=None, b=None):
        self.__r, self.__g, self.__b = map(lambda args: _ratio_or(*args), zip((r, g, b), self.__get_rgb()))

    @property
    def rgb(self) -> RGBColorProxy:
        """
        Overview:
            Get rgb color system based color proxy.
            See :class:`RGBColorProxy`.

        Returns:
            - proxy (:obj:`RGBColorProxy`): Rgb color proxy.
        """
        return RGBColorProxy(
            self,
            GetSetProxy(
                lambda: self.__r,
                lambda x: self.__set_rgb(r=_r_mapper(x)),
            ),
            GetSetProxy(
                lambda: self.__g,
                lambda x: self.__set_rgb(g=_g_mapper(x)),
            ),
            GetSetProxy(
                lambda: self.__b,
                lambda x: self.__set_rgb(b=_b_mapper(x)),
            ),
        )

    def __get_hsv(self):
        return colorsys.rgb_to_hsv(self.__r, self.__g, self.__b)

    def __set_hsv(self, h=None, s=None, v=None):
        h, s, v = map(lambda args: _ratio_or(*args), zip((h, s, v), self.__get_hsv()))
        self.__r, self.__g, self.__b = colorsys.hsv_to_rgb(h, s, v)

    @property
    def hsv(self) -> HSVColorProxy:
        """
        Overview:
            Get hsv color system based color proxy.
            See :class:`HSVColorProxy`.

        Returns:
            - proxy (:obj:`HSVColorProxy`): Hsv color proxy.
        """
        return HSVColorProxy(
            self,
            GetSetProxy(
                lambda: self.__get_hsv()[0],
                lambda x: self.__set_hsv(h=_hsv_h_mapper(x)),
            ),
            GetSetProxy(
                lambda: self.__get_hsv()[1],
                lambda x: self.__set_hsv(s=_hsv_s_mapper(x)),
            ),
            GetSetProxy(
                lambda: self.__get_hsv()[2],
                lambda x: self.__set_hsv(v=_hsv_v_mapper(x)),
            ),
        )

    def __get_hls(self):
        return colorsys.rgb_to_hls(self.__r, self.__g, self.__b)

    def __set_hls(self, h=None, l_=None, s=None):
        h, l, s = map(lambda args: _ratio_or(*args), zip((h, l_, s), self.__get_hls()))
        self.__r, self.__g, self.__b = colorsys.hls_to_rgb(h, l, s)

    @property
    def hls(self) -> HLSColorProxy:
        """
        Overview:
            Get hls color system based color proxy.
            See :class:`HLSColorProxy`.

        Returns:
            - proxy (:obj:`HLSColorProxy`): Hls color proxy.
        """
        return HLSColorProxy(
            self,
            GetSetProxy(
                lambda: self.__get_hls()[0],
                lambda x: self.__set_hls(h=_hls_h_mapper(x)),
            ),
            GetSetProxy(
                lambda: self.__get_hls()[1],
                lambda x: self.__set_hls(l_=_hls_l_mapper(x)),
            ),
            GetSetProxy(
                lambda: self.__get_hls()[2],
                lambda x: self.__set_hls(s=_hls_s_mapper(x)),
            ),
        )

    def __get_hex(self, include_alpha: bool):
        rs, gs, bs = _ratio_to_hex(self.__r), _ratio_to_hex(self.__g), _ratio_to_hex(self.__b)
        as_ = _ratio_to_hex(self.__alpha) if self.__alpha is not None and include_alpha else ''

        return '#' + rs + gs + bs + as_

    def __get_name(self):
        _hex = self.__get_hex(False).lower()
        return _CSS3_NAME_MAPS.get(_hex, _hex)

    def __repr__(self):
        if self.__alpha is not None:
            return '<{cls} {hex}, alpha: {alpha}>'.format(
                cls=self.__class__.__name__,
                hex=self.__get_name(),
                alpha='%.3f' % (self.__alpha,),
            )
        else:
            return '<{cls} {hex}>'.format(
                cls=self.__class__.__name__,
                hex=self.__get_name(),
            )

    def __str__(self):
        """
        Hex format of this :class:`Color` object.
        """
        return self.__get_hex(True)

    def __getstate__(self) -> Tuple[float, float, float, Optional[float]]:
        """
        Overview:
            Dump color as pickle object.

        Returns:
            - info (:obj:`Tuple[float, float, float, Optional[float]]`): Dumped data object.
        """
        return self.__r, self.__g, self.__b, self.__alpha

    def __setstate__(self, v: Tuple[float, float, float, Optional[float]]):
        """
        Overview:
            Load color from pickle object.

        Args:
            - v (:obj:`Tuple[float, float, float, Optional[float]]`): Dumped data object.
        """
        self.__r, self.__g, self.__b, self.__alpha = v

    def __hash__(self):
        """
        Overview:
            Get hash value of current object.

        Returns:
            - hash (:obj:`int`): Hash value of current color.
        """
        return hash(self.__getstate__())

    def __eq__(self, other):
        """
        Overview:
            Get equality between colors.

        Arguments:
            - other: Another object.

        Returns:
            - equal (:obj:`bool`): Equal or not.
        """
        if other is self:
            return True
        elif type(other) == type(self):
            return other.__getstate__() == self.__getstate__()
        else:
            return False

    @classmethod
    def from_rgb(cls, r, g, b, alpha=None) -> 'Color':
        """
        Overview:
            Load color from rgb system.

        Arguments:
            - r (:obj:`float`): Red value, should be a float value in :math:`\\left[0, 1\\right)`.
            - g (:obj:`float`): Green value, should be a float value in :math:`\\left[0, 1\\right]`.
            - b (:obj:`float`): Blue value, should be a float value in :math:`\\left[0, 1\\right]`.
            - alpha (:obj:`Optional[float]`): Alpha value, should be a float value \
                in :math:`\\left[0, 1\\right]`, default is None which means no alpha value is used.

        Returns:
            - color (:obj:`Color`): Color object.
        """
        return cls((r, g, b), alpha)

    @classmethod
    def from_hex(cls, hex_: str) -> 'Color':
        r"""
        Overview:
            Load color from hexadecimal rgb string.

        Arguments:
            - hex\_ (:obj:`str`): Hexadecimal string, maybe starts with ``#``.

        Returns:
            - color (:obj:`Color`): Color object.
        """
        return cls(hex_)

    @classmethod
    def from_hsv(cls, h, s, v, alpha=None) -> 'Color':
        """
        Overview:
            Load color from hsv system.

        Arguments:
            - h (:obj:`float`): Hue value, should be a float value in :math:`\\left[0, 1\\right)`.
            - s (:obj:`float`): Saturation value, should be a float value in :math:`\\left[0, 1\\right]`.
            - v (:obj:`float`): Brightness (value) value, should be a float value \
                in :math:`\\left[0, 1\\right]`.
            - alpha (:obj:`Optional[float]`): Alpha value, should be a float value \
                in :math:`\\left[0, 1\\right]`, default is None which means no alpha value is used.

        Returns:
            - color (:obj:`Color`): Color object.
        """
        return cls(colorsys.hsv_to_rgb(h, s, v), alpha)

    @classmethod
    def from_hls(cls, h: float, l: float, s: float, alpha: Optional[float] = None) -> 'Color':
        """
        Overview:
            Load color from hls system.

        Arguments:
            - h (:obj:`float`): Hue value, should be a float value in :math:`\\left[0, 1\\right)`.
            - l (:obj:`float`): Lightness value, should be a float value in :math:`\\left[0, 1\\right]`.
            - s (:obj:`float`): Saturation value, should be a float value in :math:`\\left[0, 1\\right]`.
            - alpha (:obj:`Optional[float]`): Alpha value, should be a float value \
                in :math:`\\left[0, 1\\right]`, default is None which means no alpha value is used.

        Returns:
            - color (:obj:`Color`): Color object.
        """
        return cls(colorsys.hls_to_rgb(h, l, s), alpha)



"""
Overview:
    Some useful utils to locate the executable files.

    Based on `hweickert/where <https://github.com/hweickert/where>`_, \
    and this version below will be long-term maintained here.
"""
import itertools
import os
from typing import Iterable, Iterator, Optional, List

from .type import is_windows

__all__ = [
    'where', 'which',
]


def where(execfile: str) -> List[str]:
    """
    Overview:
        Returns all matching file paths.

    :param execfile: Executable file to locate (such as ``python``).
    :return: The list of absolute paths of the executable files.

    Examples::
        >>> from hbutils.system import where
        >>>
        >>> where('apt-get')
        ['/usr/bin/apt-get', '/bin/apt-get']
        >>> where('bash')
        ['/usr/bin/bash', '/bin/bash']
        >>> where('not_installed')
        []
    """
    return list(_iter_where(execfile))


def which(execfile: str) -> Optional[str]:
    """
    Overview:
        Returns first matching file path, which is the one when we operate in terminal.

    :param execfile: Executable file to locate (such as ``python``).
    :return: Absolute path fo the executable file.

    Examples::
        >>> from hbutils.system import which
        >>>
        >>> which('apt-get')
        '/usr/bin/apt-get'
        >>> which('bash')
        '/usr/bin/bash'
        >>> which('not_installed')
        None
    """
    try:
        return next(_iter_where(execfile))
    except StopIteration:
        return None


def _iter_where(filename: str) -> Iterator[str]:
    """
    Like where() but returns an iterator.
    It is originally ``iwhere`` function, but now hidden because :func:`where` is the better choice than this.
    """
    possible_paths = _gen_possible_matches(filename)
    existing_file_paths = filter(_is_executable, possible_paths)
    return existing_file_paths


def _is_executable(filename: str) -> bool:
    """
    Check if the file is an executable file, which should be

    * Exist (of course)
    * Is a file (not a directory)
    * Is executable (not a common file)
    """
    return os.path.exists(filename) and os.path.isfile(filename) and os.access(filename, os.X_OK)


# os.path.normcase and os.path.normpath is VERRRRRRRRY important here,
# because the expression form of a path is actually not unique, especially on Windows.
def _normpath(filename: str) -> str:
    return os.path.normcase(os.path.normpath(os.path.abspath(filename)))


def _unique_str(siter: Iterable[str]) -> Iterator[str]:
    _exist_str = set()
    for s in siter:
        if s not in _exist_str:
            yield s


def _gen_possible_matches(filename) -> Iterator[str]:
    path_parts = os.environ.get("PATH", "").split(os.pathsep)
    if is_windows():  # Only in Windows, the executable file in current directory can be called.
        path_parts = itertools.chain((os.curdir,), path_parts)

    possible_paths = map(lambda x: os.path.join(x, filename), path_parts)
    if is_windows():
        possible_paths = itertools.chain(
            *map(lambda path: (path, f"{path}.bat", f"{path}.cmd", f"{path}.com", f"{path}.exe"), possible_paths))

    return _unique_str(map(_normpath, possible_paths))



from typing import BinaryIO, Union

from .base import CIOType

__all__ = [
    'CBufferType',
    'c_buffer',
]


class CBufferType(CIOType):
    """
    Overview:
        Bytes type.
    """

    def __init__(self, size: int):
        """
        Constructor of :class:`CBufferType`.

        :param size: Size of the buffer.
        """
        self.__size = size

    @property
    def size(self):
        """
        Size of the buffer.
        """
        return self.__size

    def read(self, file: BinaryIO) -> bytes:
        """
        Read bytes value.

        :param file: Binary file, ``io.BytesIO`` is supported as well.
        :return: Bytes value.
        """
        return file.read(self.__size)

    def write(self, file: BinaryIO, val: Union[bytearray, bytes]):
        """
        Write bytes value to binary IO object.
        If the length is not enough, ``\\\x00`` will be filled afterwards to reach the given ``size``.

        :param file: Binary file, ``io.BytesIO`` is supported as well.
        :param val: Bytes value to write.
        """
        if not isinstance(val, (bytearray, bytes)):
            raise TypeError(f'Bytearray or bytes expected, but {repr(val)}.')
        if not (0 <= len(val) <= self.__size):
            raise ValueError(f'Size is expected to be no more than {self.__size}, but actual length is {len(val)}.')

        fval = val[:self.__size]
        fval = fval + b'\x00' * (self.__size - len(fval))
        file.write(fval)


def c_buffer(size: int) -> CBufferType:
    """
    Overview:
        Reading and writing bytes with given size.

    :param size: Size of the buffer.
    :return: :class:`CBufferType` object.

    Examples::
        >>> import io
        >>> from hbutils.binary import c_buffer
        >>>
        >>> with io.BytesIO(b'\\xde\\xad\\xbe\\xef\\x00\\x12\\x34\\x56\\x78\\x00') as file:
        ...     print(c_buffer(1).read(file))
        ...     print(c_buffer(2).read(file))
        ...     print(c_buffer(3).read(file))
        ...     print(c_buffer(4).read(file))
        b'\\xde'
        b'\\xad\\xbe'
        b'\\xef\\x00\\x12'
        b'4Vx\\x00'
        >>> with io.BytesIO() as file:
        ...     c_buffer(1).write(file, b'\\xde')
        ...     c_buffer(2).write(file, b'\\xad\\xbe')
        ...     c_buffer(3).write(file, b'\\xef\\x00\\x12')
        ...     c_buffer(4).write(file, b'4Vx')  # length is 3
        ...     print(file.getvalue())
        b'\\xde\\xad\\xbe\\xef\\x00\\x124Vx\\x00'
    """
    return CBufferType(size)



"""
Overview:
    Base64 encode and decode.
"""
import base64
import warnings
from typing import Optional

__all__ = [
    'base64_encode', 'base64_decode',
]


def base64_encode(data: bytes, altchars: Optional[bytes] = None, urlsafe: bool = False) -> str:
    r"""
    Overview:
        Encoding the given binary ``data`` to base64 string.

    Arguments:
        - data (:obj:`bytes`): Binary data to be encoded.
        - altchars (:obj:`Optional[bytes]`): Characters to be altered, default is ``None``.
        - urlsafe (:obj:`bool`): Enable urlsafe mode, default is ``False``.

    Examples::

        >>> base64_encode(b'jvMIQ?K;]kNn2?1KD5H>')
        anZNSVE/Sztda05uMj8xS0Q1SD4=
        >>> base64_encode(b'jvMIQ?K;]kNn2?1KD5H>', urlsafe=True)
        anZNSVE_Sztda05uMj8xS0Q1SD4=
    """
    if urlsafe and altchars:
        warnings.warn(UserWarning('Urlsafe enabled in base64_encode, '
                                  'value altchars argument will be ignored.'), stacklevel=2)

    if urlsafe:
        return base64.urlsafe_b64encode(data).decode()
    else:
        return base64.b64encode(data, altchars).decode()


def base64_decode(base64_str: str, altchars: Optional[bytes] = None, urlsafe: bool = False) -> bytes:
    r"""
    Overview:
        Decode the given base64 string ``base64_str`` back to binary data.

    Arguments:
        - base64_str (:obj:`str`): Base64 string to be decoded.
        - altchars (:obj:`Optional[bytes]`): Characters to be altered, default is ``None``.
        - urlsafe (:obj:`bool`): Enable urlsafe mode, default is ``False``.

    Examples::

        >>> base64_decode('anZNSVE/Sztda05uMj8xS0Q1SD4=')
        b'jvMIQ?K;]kNn2?1KD5H>'
        >>> base64_decode('anZNSVE_Sztda05uMj8xS0Q1SD4=', urlsafe=True)
        b'jvMIQ?K;]kNn2?1KD5H>'
    """
    if urlsafe and altchars:
        warnings.warn(UserWarning('Urlsafe enabled in base64_decode, '
                                  'value altchars argument will be ignored.'), stacklevel=2)

    if urlsafe:
        return base64.urlsafe_b64decode(base64_str)
    else:
        return base64.b64decode(base64_str, altchars)



from .aetg import AETGGenerator
from .base import BaseGenerator
from .func import *
from .func import __all__ as _func_all
from .matrix import MatrixGenerator

__all__ = [
    'BaseGenerator', 'MatrixGenerator', 'AETGGenerator',
    *_func_all,
]



"""
Overview:
    Random function about the sequences.
"""
import random as random_module
from typing import Collection, TypeVar

__all__ = ['shuffle', 'multiple_choice']

_random_inst = getattr(random_module, '_inst')
_ElementType = TypeVar('_ElementType')


def shuffle(seq: Collection[_ElementType], *, random=None):
    """
    Overview:
        Shuffle the given collection.

    Arguments:
        - seq (:obj:`Collection[_ElementType]`): Original sequence to be shuffled.
        - random: Random instance, default is ``None`` which means use the native instance \
            from ``random`` module.

    Examples::
        >>> shuffle([1, 2, 3])
        [3, 1, 2]  # just one of the possiblities
        >>> shuffle(('a', 1, 'b', 2))
        ('b', 1, 2, 'a')  # the tuple type will be kept
    """
    random = random or _random_inst
    seq_type, seq = type(seq), list(seq)
    ids = list(range(len(seq)))
    random.shuffle(ids)

    # noinspection PyArgumentList
    return seq_type([seq[i] for i in ids])


def multiple_choice(seq: Collection[_ElementType], count: int, *,
                    put_back: bool = False, random=None):
    r"""
    Overview:
        Choose multiple items from the given sequence ``seq``.

    Arguments:
        - seq (:obj:`Collection[_ElementType]`): Original sequence to be chosen.
        - count (:obj:`int`): Choose count. Should be no more than length of ``seq`` \
            when put back is disabled.
        - put\_back (:obj:`bool`): Put back the chosen items or not, default is ``False`` \
            which means do not put back, the chosen items will be unique.
        - random: Random instance, default is ``None`` which means use the native instance \
            from ``random`` module.

    Examples::
        >>> multiple_choice([1, 2, 3], 2)
        [2, 3]  # one of the possibilities
        >>> multiple_choice([1, 2, 3], 2, put_back=True)
        [2, 2]  # this is possible when put back option is on
    """
    random = random or _random_inst
    seq_type, seq = type(seq), list(seq)
    n = len(seq)
    if not put_back and count > n:
        raise ValueError(f'Choice put back disabled, count must be no more than {repr(n)} but {repr(count)} found.')

    if put_back:
        ids = [random.randint(0, n - 1) for _ in range(count)]
    else:
        ids = shuffle(list(range(n)), random=random)[:count]

    # noinspection PyArgumentList
    return seq_type([seq[i] for i in ids])



import random
from random import _inst as _DEFAULT_RANDOM
from typing import Optional

__all__ = [
    'random_bytes',
]


def random_bytes(length: int = 32, allow_zero: bool = False, rnd: Optional[random.Random] = None) -> bytes:
    r"""
    Overview:
        Generate random bytes.

    Arguments:
        - length (:obj:`int`): Length of bytes, default is 32.
        - allow_zero (:obj:`bool`): Allow 0x00 appear in the result, default is ``False``.
        - rnd (:obj:`Optional[random.Random]`): Random object you used, \
            default is ``None`` which means just use the default one provided by system.

    Returns:
        - bytes (:obj:`bytes`): Random bytes.

    Examples::
        >>> from hbutils.random import random_bytes
        >>> random_bytes()
        b'\xdc\xdc$\x05\x83\x04\x812\xfd\xda^\x7f[{\xbc\x99\x88*\xab3\x87}\xd1\xab\xddc\xa2p\xb2\xcb\x07\xc5'
        >>> random_bytes(64)
        b"i7\x98\xd5\x81\x1d\xdb\xd8\xe1^\xf2\xe4\xbf\xe0O^\xeb\xed\xb0i\xaa\xf3\x16Jx\xf7J\xd7\xae1\x81\xc6\xad\xd21\x15\x8aX\xb6\xc7\x85\xa4\x1c{\xac^6\xdf\x03\x94kR}\x91\x96\xfe\x06{'I\xed5\x03r"
    """
    rnd = rnd or _DEFAULT_RANDOM
    bits = []
    for i in range(length):
        if allow_zero:
            bits.append(rnd.randint(0, 0xff))
        else:
            bits.append(rnd.randint(1, 0xff))

    return bytes(bits)



"""
Overview:
    Functions for file processing.
"""
import glob as gb
import os
import pathlib
from typing import Iterator

__all__ = [
    'touch', 'glob',
]


def touch(file: str, exist_ok: bool = True, makedirs: bool = True):
    """
    Overview:
        Touch the file at given path.
        Just like the ``touch`` command in unix system.

    :param file: Path of the file.
    :param exist_ok: Exist is okay or not.
    :param makedirs: Create directories when necessary.

    .. note::
        You can use this like ``touch`` command on unix.

    Examples::
        >>> import os
        >>> from hbutils.system import touch
        >>> os.listdir('.')
        []
        >>> touch('simple.txt')  # touch simple file
        >>> touch('1/2/3/simple.txt')  # touch file in nested directory (1/2/3 will be created)
        >>> os.listdir('.')
        ['1', 'simple.txt']
        >>> os.listdir('1/2/3')
        ['simple.txt']
    """
    if makedirs:
        path, _ = os.path.split(file)
        if path:
            os.makedirs(path, exist_ok=exist_ok)
    pathlib.Path(file).touch(exist_ok=exist_ok)


def glob(*items) -> Iterator[str]:
    """
    Overview:
        Glob filter by the given ``items``.

    :param items: Filter items.
    :return: Filtered existing paths.

    .. note::
        :func:`glob` is different from native ``glob.glob``, for its return value is a generator instead of list.

    Examples::
        >>> from hbutils.system import glob
        >>>
        >>> list(glob('*.md'))  # simple filter
        ['CONTRIBUTING.md', 'README.md']
        >>> list(glob('*.md', '*.txt'))  # multiple filter
        ['CONTRIBUTING.md', 'README.md', 'requirements-test.txt', 'requirements-doc.txt', 'requirements.txt']
        >>> print(*glob('hbutils/system/**/*.py'), sep=\'\\n\')  # nested filter
        hbutils/system/__init__.py
        hbutils/system/filesystem/directory.py
        hbutils/system/filesystem/file.py
        hbutils/system/filesystem/__init__.py
        hbutils/system/python/package.py
        hbutils/system/python/implementation.py
        hbutils/system/python/version.py
        hbutils/system/python/__init__.py
        hbutils/system/os/type.py
        hbutils/system/os/__init__.py
    """
    for item in items:
        yield from gb.glob(item, recursive=True)



"""
Overview:
    Base interface to quickly implement a comparable object.
"""
import operator as ops

__all__ = [
    'IComparable',
]


class IComparable:
    """
    Overview:
        Interface for a comparable object.

    Examples::
        >>> from hbutils.model import IComparable
        >>> class MyValue(IComparable):
        ...     def __init__(self, v) -> None:
        ...         self._v = v
        ...
        ...     def _cmpkey(self):
        ...         return self._v
        ...

        >>> MyValue(1) == MyValue(1)
        True
        >>> MyValue(1) == MyValue(2)
        False
        >>> MyValue(1) != MyValue(2)
        True
        >>> MyValue(1) > MyValue(2)
        False
        >>> MyValue(1) >= MyValue(2)
        False
        >>> MyValue(1) < MyValue(2)
        True
        >>> MyValue(1) <= MyValue(2)
        True
    """

    def _cmpkey(self):
        """
        Function for getting a key value which is used for comparison.

        :return: A value used to compare.
        """
        raise NotImplementedError  # pragma: no cover

    def _cmpcheck(self, op, other, default=False):
        if type(self) == type(other):
            return op(self._cmpkey(), other._cmpkey())
        else:
            return default

    def __eq__(self, other):
        if self is other:
            return True
        else:
            return self._cmpcheck(ops.__eq__, other, default=False)

    def __ne__(self, other):
        if self is other:
            return False
        else:
            return self._cmpcheck(ops.__ne__, other, default=True)

    def __lt__(self, other):
        return self._cmpcheck(ops.__lt__, other)

    def __le__(self, other):
        return self._cmpcheck(ops.__le__, other)

    def __gt__(self, other):
        return self._cmpcheck(ops.__gt__, other)

    def __ge__(self, other):
        return self._cmpcheck(ops.__ge__, other)



from functools import lru_cache
from typing import Callable

__all__ = [
    'efunc',
    'Expression',
]


@lru_cache()
def _raw_expr_func():
    return getattr(Expression, '_expr')


def _raw_expr(e) -> 'Expression':
    return _raw_expr_func()(e)


def efunc(e) -> Callable:
    """
    Overview:
        Get callable object from any types.

    :param e: Original object.
    :return: Callable object. If given ``e`` is an :class:`Expression`, its callable method will be returned. \
        If given ``e`` is a function, an equivalent method will be returned. Otherwise, a method which always return \
        ``e`` will be returned.

    .. note::
        This is the key feature of the native expressions, you need to use :func:`efunc` function to transform \
        expressions to callable functions.

    Examples::
        >>> from hbutils.expression import keep, efunc, expr
        >>>
        >>> e1 = keep()
        >>> efunc(e1 == 1)(1)
        True
        >>> efunc(e1 == 1)(2)
        False
        >>>
        >>> e2 = expr(lambda x: x + 2)
        >>> efunc(e2 == 1)(-1)
        True
        >>> efunc(e2 == 1)(1)
        False
    """
    return getattr(_raw_expr(e), '_fcall')


class Expression:
    """
    Overview:
        Base class of expressions.
    """

    def __init__(self, func=None):
        """
        Constructor of :class:`Expression`.

        :param func: Callable function, default is ``None`` which means a ``lambda x: x`` will be used.
        """
        self._fcall = func or (lambda x: x)

    def _func(self, func: Callable, *args, **kwargs):
        """
        Expression building based on given ``func`` and arguments.

        :param func: Logical function.
        :param args: Positional arguments.
        :param kwargs: Key-word arguments.
        :return: New expression with current class.

        Examples::
            >>> from hbutils.expression import efunc, Expression
            >>>
            >>> class MyExpression(Expression):
            ...     def add(self, other):
            ...         return self._func(lambda x, y: x + y, self, other)
            ...
            >>>
            >>> e1 = MyExpression()
            >>> efunc(e1.add(1))(5)  # 5 + 1 = 6
            6
            >>> efunc(e1.add(e1.add(1)))(5)  # 5 + (5 + 1) = 11
            11
        """
        _args = tuple(efunc(v) for v in args)
        _kwargs = {k: efunc(v) for k, v in kwargs.items()}

        def _new_func(x):
            return func(
                *(v(x) for v in _args),
                **{k: v(x) for k, v in _kwargs.items()},
            )

        return self.__class__(_new_func)

    @classmethod
    def _expr(cls, v):
        """
        Build expression with this class.

        :param v: Any types of value.
        :return: An expression object.
        """
        if isinstance(v, Expression):
            return v
        elif callable(v):
            return cls(v)
        else:
            return cls(lambda x: v)



import os
import shutil
import tempfile
import warnings
import weakref
from tempfile import TemporaryDirectory as _OriginTemporaryFileClass

from .requires import vpython

if vpython >= '3.8':
    TemporaryDirectory = _OriginTemporaryFileClass
else:
    class TemporaryDirectory(object):  # pragma: no cover
        """
        THIS CLASS IS COPIED FROM PYTHON3.8's TEMPFILE.
        Because PermissionError will be raised when use native TemporaryDirectory on Windows python3.7.
        This class should be removed when python3.7 is no longer supported.

        Create and return a temporary directory.  This has the same
        behavior as mkdtemp but can be used as a context manager.  For
        example:

            with TemporaryDirectory() as tmpdir:
                ...

        Upon exiting the context, the directory and everything contained
        in it are removed.
        """

        def __init__(self, suffix=None, prefix=None, dir=None):
            self.name = tempfile.mkdtemp(suffix, prefix, dir)
            self._finalizer = weakref.finalize(
                self, self._cleanup, self.name,
                warn_message="Implicitly cleaning up {!r}".format(self))

        @classmethod
        def _rmtree(cls, name):
            def onerror(func, path, exc_info):
                if issubclass(exc_info[0], PermissionError):
                    def resetperms(path):
                        try:
                            os.chflags(path, 0)
                        except AttributeError:
                            pass
                        os.chmod(path, 0o700)

                    try:
                        if path != name:
                            resetperms(os.path.dirname(path))
                        resetperms(path)

                        try:
                            os.unlink(path)
                        # PermissionError is raised on FreeBSD for directories
                        except (IsADirectoryError, PermissionError):
                            cls._rmtree(path)
                    except FileNotFoundError:
                        pass
                elif issubclass(exc_info[0], FileNotFoundError):
                    pass
                else:
                    raise

            shutil.rmtree(name, onerror=onerror)

        @classmethod
        def _cleanup(cls, name, warn_message):
            cls._rmtree(name)
            warnings.warn(warn_message, ResourceWarning)

        def __repr__(self):
            return "<{} {!r}>".format(self.__class__.__name__, self.name)

        def __enter__(self):
            return self.name

        def __exit__(self, exc, value, tb):
            self.cleanup()

        def cleanup(self):
            if self._finalizer.detach():
                self._rmtree(self.name)



from contextlib import contextmanager
from functools import wraps
from itertools import chain
from typing import Iterator, Tuple, Any, Union, List, Dict
from unittest.mock import patch, MagicMock

from hbutils.reflection import quick_import_object

__all__ = [
    'isolated_entry_points',
]


class _FakeEntryPoint:
    def __init__(self, name, dist):
        self.__name = name
        self.__dist = dist

    @property
    def name(self) -> str:
        return self.__name

    def load(self):
        return self.__dist


_max_fake_id = 0


def _fake_id() -> int:
    global _max_fake_id
    _max_fake_id += 1
    return _max_fake_id


def _fake_entry_name() -> str:
    return f'unnamed_fake_entry_{_fake_id()}'


def _yield_from_units(fes, auto_import: bool = True) -> Iterator[Tuple[str, Any]]:
    if isinstance(fes, (list, tuple)):
        for item in fes:
            if isinstance(item, tuple) and len(item) == 2:
                name, dist = item
                if auto_import and isinstance(dist, str):
                    dist, _, _ = quick_import_object(dist)
            elif auto_import and isinstance(item, str):
                dist, _, name = quick_import_object(item)
            elif hasattr(item, '__name__'):
                name, dist = item.__name__, item
            else:
                name, dist = _fake_entry_name(), item

            yield name, dist

    elif isinstance(fes, dict):
        for name, dist in fes.items():
            if auto_import and isinstance(dist, str):
                dist, _, _ = quick_import_object(dist)

            yield name, dist

    else:
        raise TypeError(f'Unknown type of fake entries - {fes!r}.')  # pragma: no cover


def _yield_fake_entries(fes, auto_import: bool = True) -> Iterator[_FakeEntryPoint]:
    for name, dist in _yield_from_units(fes, auto_import):
        yield _FakeEntryPoint(name, dist)


@contextmanager
def isolated_entry_points(group: str, fakes: Union[List, Dict[str, Any], None] = None,
                          auto_import: bool = True, clear: bool = False):
    """
    Overview:
        Isolation for :func:`pkg_resources.iter_entry_points` function.
        Can be used to fake the plugins, or just disable the installed plugins.

    :param group: Group name.
    :param fakes: Fake entry points. Dict or list are accepted.
    :param auto_import: Auto import the object from string. Default is ``True`` which means if \
        a string is given, dynamic import will be performed.
    :param clear: Clear the original entry points or not. Default is ``False`` which means the original \
        entry points will be kept and be able to be iterated.

    Examples::
        >>> import pkg_resources
        >>> from hbutils.testing import isolated_entry_points
        >>> with isolated_entry_points('my_plugin', [
        ...     (
        ...             'quick_import_object',  # name import
        ...             'hbutils.reflection.quick_import_object'
        ...     ),
        ...     ('func_filter', filter),  # named entry object
        ...     map,  # simple function
        ...     'hbutils.system.is_binary_file',  # simple import
        ... ]):
        ...     print({ep.name: ep.load() for ep in
        ...            pkg_resources.iter_entry_points('my_plugin')})
        {'quick_import_object': <function quick_import_object at 0x7fb17f4e5170>, 'func_filter': <class 'filter'>, 'map': <class 'map'>, 'is_binary_file': <function is_binary_file at 0x7fb17f55eef0>}

        >>> with isolated_entry_points('my_plugin', {
        ...     'func_map': map, 'func_binary': 'hbutils.system.is_binary_file'}):
        ...     print({ep.name: ep.load() for ep in
        ...            pkg_resources.iter_entry_points('my_plugin')})
        {'func_map': <class 'map'>, 'func_binary': <function is_binary_file at 0x7fb17f55eef0>}
    """
    if fakes is not None and not isinstance(fakes, (list, tuple, dict)):
        raise TypeError(f'Unknown entry point type - {fakes!r}.')

    from pkg_resources import iter_entry_points as _origin_iep
    group_name = group

    # noinspection PyShadowingNames
    @wraps(_origin_iep)
    def _new_iter_func(group, name=None):
        _exist_names = set()

        def _check_name(x) -> bool:
            if (name is None or x.name == name) and x.name not in _exist_names:
                _exist_names.add(x.name)
                return True
            else:
                return False

        if group == group_name:
            mocked = _yield_fake_entries(fakes or [], auto_import)
            if not clear:
                mocked = chain(mocked, _origin_iep(group, name))
            yield from filter(_check_name, mocked)
        else:
            yield from _origin_iep(group, name)

    with patch('pkg_resources.iter_entry_points', MagicMock(side_effect=_new_iter_func)):
        yield



"""
Overview:
    Useful functions for processing python functions.
"""
import warnings
from functools import wraps
from inspect import signature, Parameter
from itertools import chain
from typing import Callable, TypeVar, Union, Type, get_type_hints, Any

from ..design import SingletonMark, decolize

__all__ = [
    'fassign', 'frename', 'fcopy',
    'args_iter', 'sigsupply',
    'dynamic_call', 'static_call',
    'pre_process', 'post_process',
    'raising', 'warning_',
    'freduce',
    'get_callable_hint',
]


def fassign(**assigns):
    """
    Overview:
        Do assignments to function.

    Arguments:
        - assigns: Assignment values.

    Returns:
        - decorator: A decorator for assigning.

    Examples::
        >>> @fassign(__name__='fff')
        >>> def func(a, b):
        >>>     return a + b
    """

    def _decorator(func):
        for k, v in assigns.items():
            setattr(func, k, v)

        return func

    return _decorator


def frename(new_name: str):
    """
    Overview:
        Rename the given function.

    Arguments:
        - new_name (:obj:`str`): New name of function.

    Returns:
        - decorator: Decorator to rename the function.

    Examples::
        >>> @frename('fff')
        >>> def func(a, b):
        >>>     return a + b
    """
    return fassign(__name__=new_name)


def fcopy(func):
    """
    Overview:
        Make a copy of given function.

    Arguments:
        - func: Function to be copied.

    Returns:
        - new_func: Copied function.

    Examples::
        >>> def func(a, b):
        ...     return a + b
        >>> nfunc = fcopy(func)
        >>> nfunc(1, 2)
        3
    """

    @wraps(func)
    def _new_func(*args, **kwargs):
        return func(*args, **kwargs)

    return _new_func


def args_iter(*args, **kwargs):
    """
    Overview:
        Iterate all the arguments with index and value.
        If argument is in `args`, the index should be integer increasing from 0.
        If argument is in `kwargs`, the index should be string which meaning the argument's name.
        The numeric indices will appear before the string indices,
        and **the order of the string indices are not approved**.

    Arguments:
        - args (:obj:`Tuple[Any]`): Argument list
        - kwargs (:obj:`Dict[str, Any]`): Argument mapping

    Example:
        >>> for index, value in args_iter(1, 2, 3, a=1, b=2, c=3)):
        >>>     print(index, value)

        The output should be

        >>> 0 1
        >>> 1 2
        >>> 2 3
        >>> a 1
        >>> b 2
        >>> c 3
    """
    for _index, _item in chain(enumerate(args), sorted(kwargs.items())):
        yield _index, _item


_SIG_WRAPPED = '__sig_wrapped__'
_DYNAMIC_WRAPPED = '__dynamic_wrapped__'


def sigsupply(func, sfunc):
    """
    Overview:
        A solution for :func:`dynamic_call`. When ``func`` is a builtin function or method \
        (which means its signature can not be captured by ``inspect.signature``), the signature of \
        ``sfunc`` will take the place, and the builtin function will be able to processed properly by \
        function :func:`dynamic_call`.

    Arguments:
        - func: Original function, can be a native function or builtin function.
        - sfunc: Supplemental function, must be a native python function which has signature. \
            Its inner logic has no importance, just provide a lambda with arguments format and \
            ``None`` return.

    Examples::

        >>> dynamic_call(max)([1, 2, 3])  # no sigsupply
        ValueError: no signature found for builtin <built-in function max>
        >>> dynamic_call(sigsupply(max, lambda x: None))([1, 2, 3])  # use it as func(x) when builtin
        3
    """
    if getattr(func, _SIG_WRAPPED, None):
        return func

    try:
        signature(func, follow_wrapped=False)
    except ValueError:
        @wraps(func)
        def _new_func(*args, **kwargs):
            return func(*args, **kwargs)

        setattr(_new_func, _SIG_WRAPPED, sfunc)
        return _new_func
    else:
        return func


def _getsignature(func):
    sfunc = getattr(func, _SIG_WRAPPED, func)
    return signature(sfunc, follow_wrapped=False)


@decolize
def dynamic_call(func: Callable):
    """
    Overview:
        Decorate function to dynamic-call-supported function.

    Arguments:
        - func (:obj:`Callable`): Original function to be decorated.

    Returns:
        - new_func (:obj:`Callable`): Decorated function.

    Example:
        >>> dynamic_call(lambda x, y: x ** y)(2, 3)  # 8
        >>> dynamic_call(lambda x, y: x ** y)(2, 3, 4)  # 8, 3rd is ignored
        >>> dynamic_call(lambda x, y, t, *args: (args, (t, x, y)))(1, 2, 3, 4, 5)  # ((4, 5), (3, 1, 2))
        >>> dynamic_call(lambda x, y: (x, y))(y=2, x=1)  # (1, 2), key word supported
        >>> dynamic_call(lambda x, y, **kwargs: (kwargs, x, y))(1, k=2, y=3)  # ({'k': 2}, 1, 3)

    .. note::

        Simple :func:`dynamic_call` **can not support builtin functions because they do not have \
        python signature**. If you need to deal with builtin functions, you can use :func:`sigsupply` \
        to add a signature onto the function when necessary.

    """
    if _is_dynamic_call(func):
        return func

    enable_args, args_count = False, 0
    enable_kwargs, kwargs_set = False, set()

    for name, param in _getsignature(func).parameters.items():
        if param.kind in {Parameter.POSITIONAL_ONLY, Parameter.POSITIONAL_OR_KEYWORD}:
            args_count += 1
        if param.kind in (Parameter.KEYWORD_ONLY, Parameter.POSITIONAL_OR_KEYWORD):
            kwargs_set |= {name}
        if param.kind == Parameter.VAR_POSITIONAL:
            enable_args = True
        if param.kind == Parameter.VAR_KEYWORD:
            enable_kwargs = True

    def _get_args(*args):
        return args if enable_args else args[:args_count]

    def _get_kwargs(**kwargs):
        return kwargs if enable_kwargs else {key: value for key, value in kwargs.items() if key in kwargs_set}

    @wraps(func)
    def _new_func(*args, **kwargs):
        return func(*_get_args(*args), **_get_kwargs(**kwargs))

    setattr(_new_func, _DYNAMIC_WRAPPED, func)
    return _new_func


def _is_dynamic_call(func: Callable):
    return not not getattr(func, _DYNAMIC_WRAPPED, None)


@decolize
def static_call(func: Callable, static_ok: bool = True):
    """
    Overview:
        Static call, anti-calculation of dynamic call.

    Arguments:
        - func (:obj:`Callable`): Given dynamic function.
        - static_ok (:obj:`bool`): Allow given function to be static, default is ``True``.

    Returns:
        - static (:obj:`Callable`): Static function.
    """
    if not static_ok and not _is_dynamic_call(func):
        raise TypeError("Given callable is already static.")

    return getattr(func, _DYNAMIC_WRAPPED, func)


def pre_process(processor: Callable):
    """
    Overview:
        Pre processor for function.

    Arguments:
        - processor (:obj:`Callable`): Pre processor.

    Returns:
        - decorator (:obj:`Callable`): Function decorator

    Example:
        >>> @pre_process(lambda x, y: (-x, (x + 2) * y))
        >>> def plus(a, b):
        >>>     return a + b
        >>>
        >>> plus(1, 2)  # 5, 5 = -1 + (1 + 2) * 2
    """
    _processor = dynamic_call(processor)

    def _decorator(func):
        @wraps(func)
        def _new_func(*args, **kwargs):
            pargs = _processor(*args, **kwargs)

            if isinstance(pargs, tuple) and len(pargs) == 2 \
                    and isinstance(pargs[0], (list, tuple)) \
                    and isinstance(pargs[1], (dict,)):
                args_, kwargs_ = tuple(pargs[0]), dict(pargs[1])
            elif isinstance(pargs, (tuple, list)):
                args_, kwargs_ = tuple(pargs), {}
            elif isinstance(pargs, (dict,)):
                args_, kwargs_ = (), dict(pargs)
            else:
                args_, kwargs_ = (pargs,), {}

            return func(*args_, **kwargs_)

        return _new_func

    return _decorator


def post_process(processor: Callable):
    """
    Overview:
        Post processor for function.

    Arguments:
        - processor (:obj:`Callable`): Post processor.

    Returns:
        - decorator (:obj:`Callable`): Function decorator

    Example:
        >>> @post_process(lambda x: -x)
        >>> def plus(a, b):
        >>>     return a + b
        >>>
        >>> plus(1, 2)  # -3
    """
    processor = dynamic_call(processor)

    def _decorator(func):
        @wraps(func)
        def _new_func(*args, **kwargs):
            _result = func(*args, **kwargs)
            return processor(_result)

        return _new_func

    return _decorator


def _is_throwable(err):
    return isinstance(err, BaseException) or (isinstance(err, type) and issubclass(err, BaseException))


def _post_for_raising(ret):
    if _is_throwable(ret):
        raise ret
    else:
        return ret


def raising(func: Union[Callable, BaseException, Type[BaseException]]):
    """
    Overview:
        Decorate function with exception object return value to a raisisng function.

    Arguments:
        - func (:obj:`Union[Callable, BaseException, Type[BaseException]]`): Not decorated function or class

    Returns:
        - decorated (:obj:`Callable`): Decorated new function

    Examples:
        >>> raising(RuntimeError)()  # RuntimeError
        >>> raising(lambda x: ValueError('value error - %s' % (repr(x), )))(1)  # ValueError, value error - 1
    """
    if _is_throwable(func):
        return raising(dynamic_call(lambda: func))
    else:
        return post_process(_post_for_raising)(func)


def _is_warning(w):
    return isinstance(w, (Warning, str)) or (isinstance(w, type) and issubclass(w, Warning))


def _warn(w):
    return w() if _is_warning(w) and isinstance(w, type) and issubclass(w, Warning) else w


def _post_for_warning(ret):
    _matched = False
    if _is_warning(ret):
        _matched, _w, args_, kwargs_ = True, ret, (), {}
    elif isinstance(ret, tuple) and len(ret) >= 1 and _is_warning(ret[0]):
        _w, ret = ret[0], ret[1:]
        if len(ret) == 1:
            if isinstance(ret[0], tuple):
                _matched, args_, kwargs_ = True, ret[0], {}
            elif isinstance(ret[0], dict):
                _matched, args_, kwargs_ = True, (), ret[0]
        elif len(ret) == 2:
            if isinstance(ret[0], tuple) and isinstance(ret[1], dict):
                _matched, args_, kwargs_ = True, ret[0], ret[1]

    if not _matched:
        return ret
    else:
        # noinspection PyUnboundLocalVariable
        warnings.warn(_warn(_w), *args_, **kwargs_)


def warning_(func: Union[Callable, Warning, Type[Warning], str]):
    """
    Overview:
        Decorate function with exception object return value to a ``warning_`` function.

    Arguments:
        - func (:obj:`Union[Callable, Warning, Type[Warning], str]`): Not decorated function or class

    Returns:
        - decorated (:obj:`Callable`): Decorated new function

    Examples:
        >>> warning_(RuntimeWarning)()  # RuntimeWarning
        >>> raising(lambda x: Warning('value warning - %s' % (repr(x), )))(1)  # Warning, value warning - 1
    """
    if _is_warning(func):
        return warning_(dynamic_call(lambda: func))
    else:
        return post_process(_post_for_warning)(func)


NO_INITIAL = SingletonMark("no_initial")

_ElementType = TypeVar("_ElementType")


def freduce(init=NO_INITIAL, pass_kwargs: bool = True):
    """
    Overview:
        Make binary function be reducible.

    Arguments:
        - init (:obj:`Any`): Initial data generator or \
            initial data, default is `NO_INITIAL` which means no initial data. \
            Missing of positional arguments is forbidden.
        - pass_kwargs (:obj:`bool`): Pass kwargs into initial function and wrapped function or not, \
            default is `True` which means pass the arguments in.

    Returns:
        - decorator (:obj:`Callable`): Decorator for the original function.

    Example:
        >>> @freduce(init=0)
        >>> def plus(a, b):
        >>>     return a + b
        >>>
        >>> plus()            # 0
        >>> plus(1)           # 1
        >>> plus(1, 2)        # 3
        >>> plus(1, 2, 3, 4)  # 10
    """
    if init is NO_INITIAL:
        init_func = None
    else:
        init_func = dynamic_call(init if hasattr(init, '__call__') else (lambda: init))

    def _decorator(func):
        func = dynamic_call(func)

        @wraps(func)
        def _new_func(*args, **kwargs) -> _ElementType:
            if not pass_kwargs and kwargs:
                warnings.warn(SyntaxWarning(
                    "Key-word arguments detected but will not be passed due to the pass_kwargs setting - {kwargs}.".format(
                        kwargs=repr(kwargs))))
            kwargs = kwargs if pass_kwargs else {}

            if init_func is None:
                if not args:
                    raise SyntaxError(
                        "No less than 1 argument expected in function {func} but 0 found.".format(func=repr(func)))
                current = args[0]
                args = args[1:]
            else:
                current = init_func(**kwargs)

            for arg in args:
                current = func(current, arg, **kwargs)

            return current

        return _new_func

    return _decorator


def get_callable_hint(f: Callable):
    """
    Overview:
        Get type hint of callable.

    Arguments:
        - f (:obj:`Callable`): Callable object.

    Returns:
        - hint: Hint of the callable.

    Example:
        >>> def f1(x: float, y: str) -> int:
        >>>     pass
        >>> get_callable_hint(f1)  # Callable[[float, str], int]
        >>>
        >>> def f2(x: float, y: str, *, z: int):
        >>>     pass
        >>> get_callable_hint(f2)  # Callable[..., Any]
    """
    count, ponly = 0, True
    for key, value in signature(f).parameters.items():
        if value.kind in {Parameter.POSITIONAL_ONLY, Parameter.POSITIONAL_OR_KEYWORD}:
            count += 1
        else:
            ponly = False

    _type_hints = get_type_hints(f)
    _return_type = _type_hints.get('return', Any)
    if ponly:
        _items = [_type_hints.get(key, Any) for key in signature(f).parameters.keys()]
        return Callable[[*_items], _return_type]
    else:
        return Callable[..., _return_type]



"""
Overview:
    Isolation for directory.
"""
import os
from contextlib import contextmanager
from typing import ContextManager, Dict, Optional

from .._base import TemporaryDirectory
from ...system import copy

__all__ = [
    'isolated_directory',
]


@contextmanager
def isolated_directory(mapping: Optional[Dict[str, str]] = None) -> ContextManager:
    """
    Overview:
        Do something in an isolated directory.

    :param mapping: Mappings for the isolated directory.

    Examples::
        - Simple usage

        >>> import os
        >>> import pathlib
        >>> from hbutils.testing import isolated_directory
        >>>
        >>> with isolated_directory():
        ...     with open('file.txt', 'w') as f:
        ...         print("Line 1", file=f)
        ...         print("Line 2rd", file=f)
        ...     print(os.listdir('.'))
        ...     print(pathlib.Path('file.txt').read_text())
        ['file.txt']
        Line 1
        Line 2rd
        >>> print(os.listdir('.'))
        ['hbutils', 'README.md', 'requirements.txt', ...]


        - Mapping files and directory inside

        >>> import os
        >>> from hbutils.testing import isolated_directory
        >>>
        >>> with isolated_directory({
        ...     'ts': 'hbutils/testing',
        ...     'README.md': 'README.md',
        ... }):
        ...     print(os.listdir('.'))
        ...     print(os.listdir('ts'))
        ['README.md', 'ts']
        ['capture', 'generator', 'isolated', '__init__.py']

    """
    _original_path = os.path.abspath(os.curdir)
    with TemporaryDirectory() as dirname:
        for dst, src in (mapping or {}).items():
            dst_position = os.path.join(dirname, dst)
            os.makedirs(os.path.dirname(dst_position), exist_ok=True)
            copy(
                os.path.join(_original_path, src),
                dst_position,
            )

        try:
            os.chdir(dirname)
            yield
        finally:
            os.chdir(_original_path)



"""
Overview:
    Testing process module, include some useful utilities for unittest.
"""
from .capture import *
from .compare import *
from .generator import *
from .isolated import *
from .requires import *
from .simulate import *



from itertools import tee, chain, islice
from typing import Iterator, Iterable, Tuple, TypeVar

__all__ = [
    'nested_for', 'progressive_for',
]

_ItemType = TypeVar('_ItemType')


def nested_for(*iters: Iterable[_ItemType]) -> Iterator[Tuple[_ItemType, ...]]:
    """
    Nested for based on several iterators.

    :param iters: Iterators to build this nested for loop.
    :return: Nested for loop iteration.

    Examples::
        >>> from hbutils.reflection import nested_for
        >>> for a, r, b in nested_for(
        ...         range(1, 3),
        ...         ['a', 'b'],
        ...         map(lambda x: x ** 2, range(1, 4))
        ... ):
        >>>     print(a, r, b)
        1 a 1
        1 a 4
        1 a 9
        1 b 1
        1 b 4
        1 b 9
        2 a 1
        2 a 4
        2 a 9
        2 b 1
        2 b 4
        2 b 9
    """
    iterators = list(iters)
    n = len(iterators)

    def _recursion(deep, selections: list):
        if deep >= n:
            yield tuple(selections)
        else:
            iterators[deep], new_copy = tee(iterators[deep])
            for item in new_copy:
                selections.append(item)
                yield from _recursion(deep + 1, selections)
                selections.pop()

    yield from _recursion(0, [])


def _yield_progressive_for(iterable: Iterable[_ItemType], n: int, offset: int) -> Iterator[Tuple[_ItemType, ...]]:
    def _recursion(deep, iters, selections: list):
        if deep >= n:
            yield tuple(selections)
        else:
            iters[0], new_iter = tee(iters[0])
            if deep > 0:
                actual_iter = islice(chain(selections[-1:], new_iter), offset, None)
            else:
                actual_iter = new_iter

            while True:
                try:
                    item = next(actual_iter)
                except (StopIteration, StopAsyncIteration):
                    break

                selections.append(item)
                iter_proxy = [actual_iter]
                yield from _recursion(deep + 1, iter_proxy, selections)
                (actual_iter,) = iter_proxy
                selections.pop()

    yield from _recursion(0, [iterable], [])


def progressive_for(iterable: Iterable[_ItemType], n: int, offset: int = 1) -> Iterator[Tuple[_ItemType, ...]]:
    """
    Progressive for based on one given ``iterable``.

    :param iterable: Iterable object for this loop.
    :param n: Depth of this loop.
    :param offset: Offset of this loop, default is ``1`` which means the first value \
        in the next level will be the one after the above level.
    :returns: Progressive for loop iteration.

    Examples::
        >>> from hbutils.reflection import progressive_for
        >>> for a, b in progressive_for(range(4), 2):
        ...     print(a, b)
        0 1
        0 2
        0 3
        1 2
        1 3
        2 3
        >>> for a, b in progressive_for(range(4), 2, 0):
        ...     print(a, b)
        0 0
        0 1
        0 2
        0 3
        1 1
        1 2
        1 3
        2 2
        2 3
        3 3
        >>> for a, b, c in progressive_for(range(5), 3):
        ...     print(a, b, c)
        0 1 2
        0 1 3
        0 1 4
        0 2 3
        0 2 4
        0 3 4
        1 2 3
        1 2 4
        1 3 4
        2 3 4
        >>> for a, b, c in progressive_for(range(5), 3, 2):
        ...     print(a, b, c)
        0 2 4
    """
    if offset < 0:
        raise ValueError(f'Offset for progressive_for should not be lower than 0, but {repr(offset)} found.')

    return _yield_progressive_for(iterable, n, offset)



"""
Overview:
    Useful functions for build class models.
"""
import os
import textwrap
from typing import Optional, Iterable

from ._info import _PACKAGE_RST
from ..config.meta import __VERSION__
from ..design.singleton import SingletonMark
from ..reflection import fassign

__all__ = [
    'get_field',
    'asitems', 'visual', 'constructor', 'hasheq', 'accessor',
]

CLASS_WRAPPER_UPDATES = ()


def _cls_field_name(cls: type, name: str):
    if name.startswith('__'):
        return f'_{cls.__name__.lstrip("_")}__{name[2:]}'
    else:
        return name


_NO_DEFAULT_VALUE = SingletonMark('no_default_value')


def get_field(obj, name: str, default=_NO_DEFAULT_VALUE):
    """
    Overview:
        Get field from object.
        Private field is supported.

    :param obj: The given object.
    :param name: Field to be got.
    :param default: Default value when failed.
    :return: Field value of the field.
    """
    _field_name = _cls_field_name(type(obj), name)
    if default is _NO_DEFAULT_VALUE:
        return getattr(obj, _field_name)
    else:
        return getattr(obj, _field_name, default)


def _cls_private_prefix(cls):
    cls_name = cls.__name__.lstrip('_')
    return f'_{cls_name}__'


def _auto_get_items_from_cls(cls):
    return cls.__items__


def _auto_get_items(obj, cls_prefix=None):
    _type = type(obj)
    try:
        return _auto_get_items_from_cls(_type)
    except AttributeError:
        cls_prefix = cls_prefix or _cls_private_prefix(_type)

        items = []
        for name in dir(obj):
            if name.startswith(cls_prefix):
                items.append(name[len(cls_prefix):])

        return sorted(items)


def _get_value(self, vname: str, cls_prefix=None):
    cls_prefix = cls_prefix or _cls_private_prefix(type(self))
    try:
        return getattr(self, vname)
    except AttributeError:
        return getattr(self, cls_prefix + vname)


def asitems(items: Iterable[str]):
    """
    Overview:
        Define fields in the class level.

    Arguments:
        - items (:obj:`Iterable[str]`): Field name items.

    Returns:
        - decorator: Decorator to decorate the given class.

    Examples::
        >>> @visual()
        >>> @constructor(doc='''
        >>>     Overview:
        >>>         This is the constructor of class :class:`T`.
        >>> ''')
        >>> @asitems(['x', 'y'])
        >>> class T:
        >>>     @property
        >>>     def x(self):
        >>>         return self.__first
        >>>
        >>>     @property
        >>>     def y(self):
        >>>         return self.__second
    """

    def _decorator(cls):
        _cls_prefix = _cls_private_prefix(cls)
        cls.__items__ = list(items)
        return cls

    return _decorator


def visual(items: Optional[Iterable] = None, show_id: bool = False):
    """
    Overview:
        Decorate class to be visible by `repr`.

    Arguments:
        - items (:obj:`Optional[Iterable]`): Items to be displayed. Default is `None`, which means \
            automatically find the private fields and display them.
        - show_id (:obj:`bool`): Show hex id in representation string or not.

    Returns:
        - decorator: Decorator to decorate the given class.

    Examples::
        >>> @visual()
        >>> class T:
        ...     def __init__(self, x, y):
        ...         self.__first = x
        ...         self.__second = y
        >>> repr(T(1, 2))
        <T x: 1, y: 2>
    """

    def _decorator(cls):
        _cls_prefix = _cls_private_prefix(cls)

        def _get_items(self):
            if items is None:
                return _auto_get_items(self, _cls_prefix)
            else:
                return items

        def __repr__(self):
            f"""
            Get representation format of class {cls.__name__}.

            .. note::
                Created by {_PACKAGE_RST}, v{__VERSION__}.
            """
            str_items = []
            for item in _get_items(self):
                if isinstance(item, str):
                    _name, _repr = item, repr
                else:
                    _name, _repr = item

                _value = _get_value(self, _name, _cls_prefix)
                try:
                    _vrepr = _repr(_value)
                except ValueError:
                    pass
                else:
                    str_items.append(f'{_name}: {_vrepr}')

            sentences = []
            if show_id:
                sentences.append(hex(id(self)))
            if str_items:
                sentences.append(', '.join(str_items))
            if sentences:
                str_sentences = ' ' + ' '.join(sentences)
            else:
                str_sentences = ''
            return f'<{type(self).__name__}{str_sentences}>'

        cls.__repr__ = __repr__
        return cls

    return _decorator


_INDENT = ' ' * 4


def constructor(params: Optional[Iterable] = None, doc: Optional[str] = None):
    r"""
    Overview:
        Decorate class to create a init function.

    Arguments:
        - params (:obj:`Optional[Iterable]`): Parameters of constructor, should be a iterator of items, \
            each item should be a single string or a tuple of string and default value. Default is `None`, \
            which means no arguments.
        - doc (:obj:`Optional[str]`): Documentation of constructor function.

    Returns:
        - decorator: Decorator to decorate the given class.

    Examples::
        >>> @constructor(['x', ('y', 2)], '''
        >>>     Overview:
        >>>         This is the constructor of class :class:`T`.
        >>> ''')
        >>> class T:
        >>>     # the same as:
        >>>     # def __init__(self, x, y=2):
        >>>     #     self.__x = x
        >>>     #     self.__y = y
        >>>
        >>>     @property
        >>>     def x(self):
        >>>         return self.__first
        >>>
        >>>     @property
        >>>     def y(self):
        >>>         return self.__second
    """

    def _decorator(cls):
        _cls_prefix = _cls_private_prefix(cls)

        if params is None:
            items = _auto_get_items_from_cls(cls)
        else:
            items = params or []
        actual_items = []
        arg_items = []
        for it in items:
            if isinstance(it, str):
                itn, itv = it, _NO_DEFAULT_VALUE
            else:
                itn, itv = it

            actual_items.append((itn, itv))
            if itv is _NO_DEFAULT_VALUE:
                arg_items.append(itn)
            else:
                arg_items.append(f'{itn}={repr(itv)}')

        _init_func_str = f"""
def __init__(self, {', '.join(arg_items)}):
{os.linesep.join(f'{_INDENT}self.{_cls_prefix}{name} = {name}' for name, _ in actual_items)}
        """
        fres = {}
        exec(_init_func_str, fres)

        _init_func = fres['__init__']
        _init_func = fassign(__doc__=doc or textwrap.dedent(f'''
        Constructor of class {cls.__name__}.

        .. note::
            Created by {_PACKAGE_RST}, v{__VERSION__}.

            The values of arguments supplied to this constructor \
            will be put into the fields specified.
        '''))(_init_func)
        cls.__init__ = _init_func

        return cls

    return _decorator


def hasheq(items: Optional[Iterable] = None):
    """
    Overview:
        Decorate class to be visible by `repr`.

    Arguments:
        - items (:obj:`Optional[Iterable]`): Items to be hashed and compared. Default is `None`, \
            which means automatically find the private fields and display them.

    Returns:
        - decorator: Decorator to decorate the given class.

    Examples::
        >>> @hasheq(['x', 'y'])
        >>> class T:
        >>>     def __init__(self, x, y):
        >>>         self.__first = x
        >>>         self.__second = y

        Using with :func:`asitems`

        >>> @hasheq()
        >>> @constructor()
        >>> @asitems(['x', 'y'])
        >>> class T1:
        >>>     pass
    """

    def _decorator(cls):
        _cls_prefix = _cls_private_prefix(cls)

        def _get_items(self):
            if items is None:
                return _auto_get_items(self, _cls_prefix)
            else:
                return items

        def _get_obj_values(self):
            return tuple(_get_value(self, name, _cls_prefix) for name in _get_items(self))

        def __hash__(self):
            f"""
            Hash value of class {cls.__name__}'s instances.

            .. note::
                Created by {_PACKAGE_RST}, v{__VERSION__}.

                The return value is calculated based on the \
                values of the internally specified fields.
            """
            return hash(_get_obj_values(self))

        def __eq__(self, other):
            f"""
            Equality between class {cls.__name__}'s instances.

            .. note::
                Created by {_PACKAGE_RST}, v{__VERSION__}.

                Currently, the return value is True only if both objects \\
                are of class {cls.__name__} (subclasses are not permitted) and the \\
                values of the internally specified fields are all equal; \\
                otherwise, it is always False.
            """
            if self is other:
                return True
            elif type(self) == type(other):
                return _get_obj_values(self) == _get_obj_values(other)
            else:
                return False

        def __ne__(self, other):
            f"""
            Non-equality between class {cls.__name__}'s instances.

            .. note::
                Created by {_PACKAGE_RST}, v{__VERSION__}.

                Currently, the return value is False only if both objects \\
                are of class {cls.__name__} (subclasses are not permitted) and the \\
                values of the internally specified fields are all equal; \\
                otherwise, it is always True.
            """
            return not __eq__(self, other)

        cls.__hash__ = __hash__
        cls.__eq__ = __eq__
        cls.__ne__ = __ne__

        return cls

    return _decorator


_READONLY_MARKS = {'r', 'ro', 'readonly'}
_WRITABLE_MARKS = {'w', 'rw', 'writable'}


def _is_writable(mark: str):
    if mark.lower() in _READONLY_MARKS:
        return False
    elif mark.lower() in _WRITABLE_MARKS:
        return True
    else:
        raise ValueError(f'Unknown accessible mark - {repr(mark)}.')


def accessor(items: Optional[Iterable] = None, readonly: bool = False):
    """
    Overview:
        Decorate class to be accessible by the accessors.

    Arguments:
        - items (:obj:`Optional[Iterable]`): Items to be hashed and compared. Default is `None`, \
            which means automatically find the private fields and display them.
        - readonly (:obj:`bool`): Default readonly or not. Default is `False`, which means make \
            the accessor be writable when ``rw`` option is not given.

    Returns:
        - decorator: Decorator to decorate the given class.

    Examples::
        >>> @accessor(readonly=True)
        >>> @asitems(['x', 'y'])
        >>> class T:
        >>>     def __init__(self, x, y):
        >>>         self.__first = x
        >>>         self.__second = y
        >>>
        >>> t = T(2, 100)
        >>> t.x  # 2
        >>> t.y  # 100
    """

    def _decorator(cls):
        _cls_prefix = _cls_private_prefix(cls)

        if items is None:
            actual_items = _auto_get_items_from_cls(cls)
        elif isinstance(items, str):
            actual_items = [items]
        else:
            actual_items = items or []

        pitems = []
        for it in actual_items:
            if isinstance(it, str):
                itn, itv = it, ('ro' if readonly else 'rw')
            else:
                itn, itv = it

            itv = _is_writable(itv)
            pitems.append((itn, itv))

        for itn, itv in pitems:
            _getter_func_str = f"""
def get_{itn}(self):
{_INDENT}return self.{_cls_prefix}{itn}
            """
            fres = {}
            exec(_getter_func_str, fres)
            getter_func = fres[f'get_{itn}']
            getter_func = fassign(__doc__=f"""
            Property {itn}.

            .. note::
                Created by {_PACKAGE_RST}, v{__VERSION__}.

                {'Both reading and writing are' if itv else 'Only reading is'} \\
                permitted with the accessor ``{itn}``.
            """)(getter_func)

            if itv:
                _setter_func_str = f"""
def set_{itn}(self, new_value):
{_INDENT}self.{_cls_prefix}{itn} = new_value
                """
                fres = {}
                exec(_setter_func_str, fres)
                setter_func = fres[f'set_{itn}']

                p = property(getter_func, setter_func)
            else:
                p = property(getter_func)

            setattr(cls, itn, p)

        return cls

    return _decorator



from operator import lt, le, gt, ge

from pkg_resources import parse_version

_Version = type(parse_version('0.0.1'))


class VersionInfo:
    """
    Overview:
        Class for wrapping version information.

    .. warning::
        This class is not immutable for its designing for dynamic comparison and boolean check.
        Please pay attention when use it.
    """

    def __init__(self, v):
        if isinstance(v, VersionInfo):
            self._version, self._func = v._version, None
        elif isinstance(v, _Version) or v is None:
            self._version, self._func = v, None
        elif callable(v):
            self._version, self._func = None, v
        elif isinstance(v, str):
            VersionInfo.__init__(self, parse_version(v))
        elif isinstance(v, tuple):
            VersionInfo.__init__(self, '.'.join(map(str, v)))
        elif isinstance(v, int):
            VersionInfo.__init__(self, str(v))
        else:
            raise TypeError(f'Unknown version type - {repr(v)}.')

    @property
    def _actual_version(self):
        if self._func is None:
            return self._version
        else:
            return VersionInfo(self._func())._version

    def _cmp(self, cmp, other):
        other = VersionInfo(other)
        if self and other:
            return cmp(self._actual_version, VersionInfo(other)._actual_version)
        else:
            return False

    def __lt__(self, other):
        return self._cmp(lt, other)

    def __le__(self, other):
        return self._cmp(le, other)

    def __gt__(self, other):
        return self._cmp(gt, other)

    def __ge__(self, other):
        return self._cmp(ge, other)

    def __eq__(self, other):
        return self._actual_version == VersionInfo(other)._actual_version

    def __ne__(self, other):
        return self._actual_version != VersionInfo(other)._actual_version

    def __bool__(self):
        return bool(self._actual_version)

    def __repr__(self):
        return f'<{type(self).__name__} {self._actual_version}>'



import io
import random
from datetime import datetime
from functools import partial
from random import _inst as _DEFAULT_RANDOM
from typing import Optional

from .binary import random_bytes
from ..encoding import md5, sha1, base64_encode

__all__ = [
    'random_digits', 'random_bin_digits', 'random_hex_digits',
    'random_md5', 'random_sha1', 'random_base64',
    'random_md5_with_timestamp', 'random_sha1_with_timestamp',
]


def _check_base(base: int):
    if base < 2:
        raise ValueError(f'Base should be an integer no less than 2, but {repr(base)} found.')
    elif base > 36:
        raise ValueError(f'Base should be an integer no greater then 36, but {repr(base)} found.')
    elif not isinstance(base, int):
        raise TypeError(f'Base should be an integer, but {repr(type(base))} found.')


_0_ASCII = ord('0')
_LOWER_A_ASCII = ord('a')
_UPPER_A_ASCII = ord('A')


def _random_dchar(base: int, upper: bool, rnd: random.Random):
    _val = rnd.randint(0, base - 1)
    if _val < 10:
        _base = _0_ASCII
    elif upper:
        _base = _UPPER_A_ASCII - 10
    else:
        _base = _LOWER_A_ASCII - 10

    return chr(_base + _val)


def random_digits(length: int = 32, base: int = 10, upper: bool = False, rnd: Optional[random.Random] = None) -> str:
    """
    Overview:
        Create random digits.

    Arguments:
        - length (:obj:`int`): Length of the digits, default is 32.
        - base (:obj:`int`): Base of the digits, should be in [2, 36], default is 10.
        - upper (:obj:`bool`): Upper the hex chars, default is ``False``.
        - rnd (:obj:`Optional[random.Random]`): Random object you used, \
            default is ``None`` which means just use the default one provided by system.

    Returns:
        - string (:obj:`str`): Random digital string.

    Examples::
        >>> from hbutils.random import random_digits
        >>> random_digits()
        '53518555004529024184262875530824'
        >>> random_digits(base=8)
        '77337055655313664176450107031511'
        >>> random_digits(48, base=8)
        '130107050101775254773050732461131017135371516420'
    """
    _check_base(base)
    rnd = rnd or _DEFAULT_RANDOM

    with io.StringIO() as sio:
        for i in range(length):
            sio.write(_random_dchar(base, upper, rnd))

        return sio.getvalue()


def random_bin_digits(length: int = 32, rnd: Optional[random.Random] = None) -> str:
    """
    Overview:
        Create random binary digits.

    Arguments:
        - length (:obj:`int`): Length of the digits, default is 32.
        - rnd (:obj:`Optional[random.Random]`): Random object you used, \
            default is ``None`` which means just use the default one provided by system.

    Returns:
        - string (:obj:`str`): Random binary digital string.

    Examples::
        >>> from hbutils.random import random_bin_digits
        >>> random_bin_digits()
        '11001011010101101100011010010011'
        >>> random_bin_digits(48)
        '010110000110101101111111011100010011101011010100'
    """
    return random_digits(length, 2, False, rnd)


def random_hex_digits(length: int = 32, upper: bool = False, rnd: Optional[random.Random] = None) -> str:
    """
    Overview:
        Create random hexidecimal digits.

    Arguments:
        - length (:obj:`int`): Length of the digits, default is 32.
        - rnd (:obj:`Optional[random.Random]`): Random object you used, \
            default is ``None`` which means just use the default one provided by system.

    Returns:
        - string (:obj:`str`): Random hexidecimal digital string.

    Examples::
        >>> from hbutils.random import random_hex_digits
        >>> random_hex_digits()
        'bf4eadfb8c1700d74024833c3ce211a7'
        >>> random_hex_digits(upper=True)
        '7B85DE69A319BA132ACA27C7777A1C3E'
        >>> random_hex_digits(48)
        '7175a23730391687b7b5230c72d702a1664833a1c66783cc'
    """
    return random_digits(length, 16, upper, rnd)


_RANDOM_BYTES_LENGTH = 64


def _random_hash(hash_func, length: int = _RANDOM_BYTES_LENGTH, rnd: Optional[random.Random] = None):
    return hash_func(random_bytes(length, allow_zero=False, rnd=rnd))


def random_md5(rnd: Optional[random.Random] = None) -> str:
    """
    Overview:
        Create random md5 string.

    Arguments:
        - rnd (:obj:`Optional[random.Random]`): Random object you used, \
            default is ``None`` which means just use the default one provided by system.

    Returns:
        - string (:obj:`str`): Random md5 string.

    Examples::
        >>> from hbutils.random import random_md5
        >>> random_md5()
        'bbffd8913a7c49161ebe31b9092a9016'
    """
    return _random_hash(md5, _RANDOM_BYTES_LENGTH, rnd)


def random_sha1(rnd: Optional[random.Random] = None) -> str:
    """
    Overview:
        Create random sha1 string.

    Arguments:
        - rnd (:obj:`Optional[random.Random]`): Random object you used, \
            default is ``None`` which means just use the default one provided by system.

    Returns:
        - string (:obj:`str`): Random sha1 string.

    Examples::
        >>> from hbutils.random import random_sha1
        >>> random_sha1()
        '13135aa6b05482dcdbc1f5a25d117298571e7fab'
    """
    return _random_hash(sha1, _RANDOM_BYTES_LENGTH, rnd)


def random_base64(length: int = _RANDOM_BYTES_LENGTH, rnd: Optional[random.Random] = None) -> str:
    """
    Overview:
        Create random base64, may be useful when matrix verification code.

    Arguments:
        - length (:obj:`int`): Length of the original binary data, default is 64.
        - rnd (:obj:`Optional[random.Random]`): Random object you used, \
            default is ``None`` which means just use the default one provided by system.

    Returns:
        - string (:obj:`str`): Random base64 string.

    Examples::
        >>> from hbutils.random import random_base64
        >>> random_base64()
        'PJZzHkM2-DpeXn1W9b3rp0I66MnOeD-31d2XYTA3va7N8DSNmQgvIINnvDMKWaRW-WHo_ftgKHg40z7XbDupbg=='
        >>> random_base64(48)
        'siRZNSeytSUXlIgKYuZOzbhehhI7oabcxFDB07PkjyZ5b0DI5hGC0pqjJFlD6NGQ'
    """
    return _random_hash(partial(base64_encode, urlsafe=True), length, rnd)


def _timestamp():
    return datetime.now().strftime("%Y%m%d%H%M%S%f")


def random_md5_with_timestamp(rnd: Optional[random.Random] = None) -> str:
    """
    Overview:
        Create random md5 string with timestamp.

    Arguments:
        - rnd (:obj:`Optional[random.Random]`): Random object you used, \
            default is ``None`` which means just use the default one provided by system.

    Returns:
        - string (:obj:`str`): Random md5 string with timestamp.

    Examples::
        >>> from hbutils.random import random_md5_with_timestamp
        >>> random_md5_with_timestamp()
        '20220116233104357175_daf67fde4b758ff4aae21cc77f5ed689'
    """
    return f'{_timestamp()}_{random_md5(rnd)}'


def random_sha1_with_timestamp(rnd: Optional[random.Random] = None) -> str:
    """
    Overview:
        Create random sha1 string with timestamp.

    Arguments:
        - rnd (:obj:`Optional[random.Random]`): Random object you used, \
            default is ``None`` which means just use the default one provided by system.

    Returns:
        - string (:obj:`str`): Random sha1 string with timestamp.

    Examples::
        >>> from hbutils.random import random_sha1_with_timestamp
        >>> random_sha1_with_timestamp()
        '20220116233121916685_fba840b80163b55cd2295d84286a438bf8acb7c0'
    """
    return f'{_timestamp()}_{random_sha1(rnd)}'



"""
Overview:
    Design pattern module. Even in python, design patterns are not emphasized like java, \
    but some simple packages are often useful, such as singleton patterns.
"""
from .decorator import *
from .final import *
from .observer import *
from .singleton import *



from typing import TypeVar, Type, List, Callable, Optional, Any, Dict

__all__ = [
    'BaseRecovery',
    'DictRecovery', 'ListRecovery', 'TupleRecovery',
    'NullRecovery', 'GenericObjectRecovery',
    'register_recovery', 'get_recovery_func',
]

_OriginType = TypeVar('_OriginType')


class BaseRecovery:
    __rtype__ = object

    def __init__(self, origin: _OriginType):
        """
        Constructor of :class:`BaseRecovery`.

        :param origin: Origin object to be recovered.
        """
        self.origin = origin

    def _recover(self):
        """
        Implementation for recovery.
        """
        raise NotImplementedError  # pragma: no cover

    def recover(self) -> _OriginType:
        """
        Recover the given object.

        :return: Recovered object.
        """
        self._recover()
        return self.origin

    @classmethod
    def _recover_child(cls, child):
        """
        Get recovered child-level object.

        :param child: Child object, should be a :class:`BaseRecovery` or native object.
        :return: Recovered child-level object.
        """
        if isinstance(child, BaseRecovery):
            return child.recover()
        else:
            return child

    @classmethod
    def from_origin(cls, origin: _OriginType, recursive: bool = True) -> 'BaseRecovery':
        """
        Create a recovery object by the given original object.

        :param origin: Original object to recover.
        :param recursive: Recursive or not. Default is ``True`` which means the child-level object \
            contained in ``origin`` will be recovered as well.
        :return: Recovery object.
        """
        raise NotImplementedError  # pragma: no cover

    @classmethod
    def _create_child(cls, child, recursive: bool = True):
        """
        Create child-level object for storage usage.

        :param child: Original child-level object.
        :param recursive: Recursive or not. Default is ``True`` which means the child-level object \
            contained in ``origin`` will be recovered as well.
        :return: Object for storage.
        """
        if recursive:
            clazz = _get_recovery_class(child)
            if clazz is not None:
                return clazz.from_origin(child, recursive)

        return child


_REC_CLASSES: Optional[List[Type[BaseRecovery]]] = None


def register_recovery(cls: Type[BaseRecovery]):
    """
    Overview:
        Register recovery class.

    :param cls: Recovery class.

    .. note::
        This API is used for customize recovery for other classes. \
        For more details, you may take a look at the source code of :class:`BaseRecovery`.
    """
    _REC_CLASSES.append(cls)


_DictType = TypeVar('_DictType', bound=dict)


class DictRecovery(BaseRecovery):
    __rtype__ = dict

    def __init__(self, origin: _DictType, mp: Dict):
        """
        Constructor of :class:`DictRecovery`.

        :param origin: Origin object to be recovered.
        """
        BaseRecovery.__init__(self, origin)
        self.mapping = mp

    def _recover(self):
        target = {
            key: self._recover_child(value)
            for key, value in self.mapping.items()
        }
        keys = set(self.origin.keys()) | set(target.keys())
        for key in keys:
            if key not in target:
                del self.origin[key]
            else:
                self.origin[key] = target[key]

    @classmethod
    def from_origin(cls, origin: _DictType, recursive: bool = True) -> 'DictRecovery':
        return cls(origin, {key: cls._create_child(value, recursive) for key, value in origin.items()})


_TupleType = TypeVar('_TupleType', bound=tuple)


class TupleRecovery(BaseRecovery):
    __rtype__ = tuple

    def __init__(self, origin: _TupleType, items: List[Any]):
        """
        Constructor of :class:`TupleRecovery`.

        :param origin: Origin object to be recovered.
        """
        BaseRecovery.__init__(self, origin)
        self.items = items

    def _recover(self):
        for item in self.items:
            self._recover_child(item)

    @classmethod
    def from_origin(cls, origin: _TupleType, recursive: bool = True) -> 'TupleRecovery':
        return cls(origin, [cls._create_child(item, recursive) for item in origin])


_ListType = TypeVar('_ListType', bound=list)


class ListRecovery(BaseRecovery):
    __rtype__ = list

    def __init__(self, origin: _ListType, items: List):
        """
        Constructor of :class:`ListRecovery`.

        :param origin: Origin object to be recovered.
        """
        BaseRecovery.__init__(self, origin)
        self.items = items

    def _recover(self):
        target = [
            self._recover_child(item)
            for item in self.items
        ]
        self.origin[:] = target

    @classmethod
    def from_origin(cls, origin: _ListType, recursive: bool = True) -> 'ListRecovery':
        return cls(origin, [cls._create_child(item, recursive) for item in origin])


class NullRecovery(BaseRecovery):
    """
    Overview:
        Empty recovery class for builtin immutable types.
    """
    __rtype__ = (int, float, str, bool, bytes, complex, range, slice)

    def _recover(self):
        """
        Just do nothing.
        """
        pass

    @classmethod
    def from_origin(cls, origin: _OriginType, recursive: bool = True) -> 'NullRecovery':
        """
        Just do nothing.
        """
        return cls(origin)


class GenericObjectRecovery(BaseRecovery):
    """
    Overview:
        Recovery class for generic objects.
        The ``__dict__`` will be recovered.

        .. note::
            If what you need to recover is not only ``__dict__``, may be you need to custom \
            recovery class by inheriting :class:`BaseRecovery` class, and register it by \
            :func:`register_recovery` function.
    """
    __rtype__ = object

    def __init__(self, origin: _OriginType, dict_: Optional['DictRecovery']):
        """
        Constructor of :class:`GenericObjectRecovery`.

        :param origin: Original object to recover.
        :param dict_: Recovery object of ``__dict__``, ``None`` when ``origin`` do not have ``__dict__``
        """
        BaseRecovery.__init__(self, origin)
        self.dict_ = dict_

    def _recover(self):
        """
        Recover the ``__dict__``.
        """
        if self.dict_ is not None:
            self.dict_.recover()

    @classmethod
    def from_origin(cls, origin: _OriginType, recursive: bool = True) -> 'BaseRecovery':
        """
        Create recovery object.
        """
        dict_ = DictRecovery.from_origin(origin.__dict__, recursive) if hasattr(origin, '__dict__') else None
        return cls(origin, dict_)


if _REC_CLASSES is None:
    _REC_CLASSES = []
    register_recovery(GenericObjectRecovery)
    register_recovery(NullRecovery)
    register_recovery(TupleRecovery)
    register_recovery(DictRecovery)
    register_recovery(ListRecovery)


def _get_recovery_class(origin: _OriginType) -> Optional[Type[BaseRecovery]]:
    for cls in reversed(_REC_CLASSES):
        if isinstance(origin, cls.__rtype__):
            return cls

    assert False, f'The object cannot be wrapped by recoveries - {origin!r}'  # pragma: no cover


def get_recovery_func(origin: _OriginType, recursive: bool = True) -> Callable[[], _OriginType]:
    """
    Overview:
        Get recovery function for given object.
        Dict, list and tuple object are natively supported.

    :param origin: Original object to recover.
    :param recursive: Recursive or not. Default is ``True`` which means the child-level object \
        contained in ``origin`` will be recovered as well.
    :return: Recovery function.

    Examples::
        >>> from hbutils.collection import get_recovery_func
        >>> l = [1, {'a': 1, 'b': 2}, 3, 4, 5]
        >>> print(id(l), l)
        140146367304720 [1, {'a': 1, 'b': 2}, 3, 4, 5]
        >>> f = get_recovery_func(l)

        >>> l[3] = 1
        >>> l.pop()
        >>> l.append('sdklfj')
        >>> l.append('sdkfhjksd')
        >>> l[1]['c'] = 2
        >>> l[1]['a'] = 100
        >>> print(id(l), l)
        140146367304720 [1, {'a': 100, 'b': 2, 'c': 2}, 3, 1, 'sdklfj', 'sdkfhjksd']

        >>> lx = f()
        >>> print(id(lx), lx)  # lx is l
        140146367304720 [1, {'a': 1, 'b': 2}, 3, 4, 5]
        >>> print(id(l), l)  # the value is recovered
        140146367304720 [1, {'a': 1, 'b': 2}, 3, 4, 5]

    """
    cls = _get_recovery_class(origin)
    return cls.from_origin(origin, recursive).recover



import ctypes
from typing import BinaryIO, Dict, List, Union

from .base import CMarkedType

__all__ = [
    'CFloatType',
    'c_float16', 'c_float32', 'c_float64',
    'c_float', 'c_double',
]


class CFloatType(CMarkedType):
    """
    Overview:
        Float type, based on ``struct`` module.
    """

    def read(self, file: BinaryIO) -> float:
        return super().read(file)

    def write(self, file: BinaryIO, val: Union[int, float]):
        super().write(file, float(val))


c_float16 = CFloatType('e', 2)
"""
Overview:
    Reading and writing half-precision(16-bits) float.
"""
c_float32 = CFloatType('f', 4)
"""
Overview:
    Reading and writing single-precision(32-bits) float.

Examples::
    >>> import io
    >>> import math
    >>> from hbutils.binary import c_float32
    >>>
    >>> with io.BytesIO(b'\\x00\\x00\\x90\\x7f'
    ...                 b'\\x00\\x00\\x80\\x7f'
    ...                 b'\\x00\\xa0\\x3e\\xc1'
    ...                 b'\\x00\\x00\\x70\\x00') as file:
    ...     print(c_float32.read(file))
    ...     print(c_float32.read(file))
    ...     print(c_float32.read(file))
    ...     print(c_float32.read(file))
    nan
    inf
    -11.9140625
    1.0285575569695016e-38
    >>> with io.BytesIO() as file:
    ...     c_float32.write(file, math.nan)
    ...     c_float32.write(file, +math.inf)
    ...     c_float32.write(file, -11.9140625)
    ...     c_float32.write(file, 1.0285575569695016e-38)
    ...     print(file.getvalue())
    b'\\x00\\x00\\xc0\\x7f\\x00\\x00\\x80\\x7f\\x00\\xa0>\\xc1\\x00\\x00p\\x00'
"""
c_float64 = CFloatType('d', 8)
"""
Overview:
    Reading and writing double-precision(64-bits) float.

Examples::
    >>> import io
    >>> import math
    >>> from hbutils.binary import c_float64
    >>>
    >>> with io.BytesIO(b'\\x00\\x00\\x00\\x00\\x00\\x00\\xf8\\x7f'
    ...                 b'\\x00\\x00\\x00\\x00\\x00\\x00\\xf0\\x7f'
    ...                 b'\\x00\\x00\\x00\\x00\\x00\\xd4\\x27\\xc0'
    ...                 b'\\x00\\x00\\x00\\x00\\x00\\x00\\x0c8') as file:
    ...     print(c_float64.read(file))
    ...     print(c_float64.read(file))
    ...     print(c_float64.read(file))
    ...     print(c_float64.read(file))
    nan
    inf
    -11.9140625
    1.0285575569695016e-38
    >>> with io.BytesIO() as file:
    ...     c_float64.write(file, math.nan)
    ...     c_float64.write(file, +math.inf)
    ...     c_float64.write(file, -11.9140625)
    ...     c_float64.write(file, 1.0285575569695016e-38)
    ...     print(file.getvalue())
    b"\\x00\\x00\\x00\\x00\\x00\\x00\\xf8\\x7f\\x00\\x00\\x00\\x00\\x00\\x00\\xf0\\x7f\\x00\\x00\\x00\\x00\\x00\\xd4'\\xc0\\x00\\x00\\x00\\x00\\x00\\x00\\x0c8"
"""

_EXIST_TYPES: List[CFloatType] = [
    c_float16,
    c_float32,
    c_float64,
]
_SIZE_TO_FLOAT_TYPE: Dict[int, CFloatType] = {
    item.size: item
    for item in _EXIST_TYPES
}


def _get_from_raw(tp) -> CFloatType:
    return _SIZE_TO_FLOAT_TYPE[ctypes.sizeof(tp)]


c_float = _get_from_raw(ctypes.c_float)
"""
Alias for :data:`c_float32`.
"""
c_double = _get_from_raw(ctypes.c_double)
"""
Alias for :data:`c_float64`.
"""



from typing import Optional, Type

from .base import Expression
from .feature import ComparableExpression, IndexedExpression, LogicalExpression, MathExpression, ObjectExpression

__all__ = [
    'GeneralExpression',
    'expr', 'keep', 'raw',
]


class GeneralExpression(
    ComparableExpression,
    IndexedExpression,
    ObjectExpression,
    LogicalExpression,
    MathExpression,
):
    """
    Overview:
        General expression.

    Features:
        Inherited from :class:`ComparableExpression`, :class:`IndexedExpression`, :class:`ObjectExpression`, \
        :class:`LogicalExpression` and :class:`MathExpression`.
    """
    pass


def expr(v, cls: Optional[Type[Expression]] = None):
    """
    Overview:
        Transform any objects to :class:`Expression`.

        * If the given ``v`` is an :class:`Expression`, ``v`` will be returned.
        * If the given ``v`` is a callable object, expression with ``v`` will be returned.
        * Otherwise, expression with ``lambda x: v`` will be returned.

    :param v: Original value.
    :param cls: Class of expression, should be a subclass of :class:`Expression`. Default is ``None`` which \
        means :class:`GeneralExpression` will be used.
    :return: Generated expression.
    """
    cls = cls or GeneralExpression
    return getattr(cls, '_expr')(v)


def keep(cls: Optional[Type[Expression]] = None):
    """
    Overview:
        Alias for ``expr(lambda x: x, cls)``.

    :param cls: Class of expression, should be a subclass of :class:`Expression`. Default is ``None`` which \
        means :class:`GeneralExpression` will be used.
    :return: Generated expression.
    """
    return expr(lambda x: x, cls)


def raw(v, cls: Optional[Type[Expression]] = None):
    """
    Overview:
        Alias for ``expr(lambda x: v, cls)``.

    :param v: Raw value.
    :param cls: Class of expression, should be a subclass of :class:`Expression`. Default is ``None`` which \
        means :class:`GeneralExpression` will be used.
    :return: Generated expression.
    """
    return expr(lambda x: v, cls)



import sys
import types
from contextlib import contextmanager
from typing import ContextManager, Mapping, List, Dict

__all__ = [
    'mount_pythonpath',
    'PythonPathEnv',
]


def _copy_list(origin: list, target: list):
    origin[:] = target


def _copy_dict(origin: dict, target: dict):
    for key in set(origin.keys()) | set(target.keys()):
        if key not in target:
            del origin[key]
        else:
            origin[key] = target[key]


@contextmanager
def _native_mount_pythonpath(paths: List[str], modules: Dict[str, types.ModuleType]) -> ContextManager:
    from ..collection import get_recovery_func
    path_rec = get_recovery_func(sys.path, recursive=False)
    modules_rec = get_recovery_func(sys.modules, recursive=False)
    try:
        _copy_list(sys.path, paths)
        _copy_dict(sys.modules, modules)
        yield
    finally:
        path_rec()
        modules_rec()


class PythonPathEnv:
    """
    Overview:
        Python environment object.
    """

    def __init__(self, pythonpath: List[str], modules: Mapping[str, types.ModuleType]):
        """
        Constructor of :class:`PythonPathEnv`.

        :param pythonpath: Python path list.
        :param modules: Modules loaded.
        """
        self.pythonpath: List[str] = list(pythonpath)
        self.modules: Dict[str, types.ModuleType] = dict(modules)

    @contextmanager
    def mount(self, keep: bool = True) -> ContextManager['PythonPathEnv']:
        """
        Mount the ``PYTHONPATH`` and modules of this environment.

        :param keep: Keep the changes inside. Default is ``True`` which means the new imports and modules \
            will be kept inside and will be usable when next time the :meth:`mount` is called.

        Examples::
            >>> from hbutils.reflection import mount_pythonpath
            >>> with mount_pythonpath('test/testfile/igm') as env:
            ...     from gf1 import FIXED
            ...     print('FIXED in igm:', FIXED)
            FIXED in igm: 1234567
            >>> with env.mount():
            ...     from gf1 import FIXED
            ...     print('FIXED in igm:', FIXED)
            FIXED in igm: 1234567
        """
        if keep:
            pythonpath, modules = self.pythonpath, self.modules
        else:
            pythonpath, modules = [*self.pythonpath], {**self.modules}

        with _native_mount_pythonpath(pythonpath, modules):
            yield self


@contextmanager
def mount_pythonpath(*path) -> ContextManager[PythonPathEnv]:
    """
    Overview:
        Append ``PYTHONPATH`` in context, the packages in given paths will be able to be imported.
        ``sys.modules`` will also be recovered when quit.

    :param path: Appended python path.
    
    Examples::
        Here is the testfile directory

        >>> import os
        >>> os.system('tree test/testfile')
        test/testfile
        ├── dir1
        │   ├── gf1.py
        ├── dir2
        │   ├── gf1.py
        └── igm
            └── gf1.py

        We can import the values from the other directories, like this

        >>> from hbutils.reflection import mount_pythonpath
        >>> with mount_pythonpath('test/testfile/igm'):
        ...     from gf1 import FIXED
        ...     print('FIXED in igm:', FIXED)
        FIXED in igm: 1234567
        >>>
        >>> with mount_pythonpath('test/testfile/dir1'):
        ...     from gf1 import FIXED
        ...     print('FIXED in dir1:', FIXED)
        FIXED in dir1: 233
        >>>
        >>> with mount_pythonpath('test/testfile/dir2'):
        ...     from gf1 import FIXED
        ...     print('FIXED in dir2:', FIXED)
        FIXED in dir2: 455
        >>>
        >>> from gf1 import FIXED  # cannot import outside
        ModuleNotFoundError: No module named 'gf1'

    """
    with _native_mount_pythonpath([*path, *sys.path], {**sys.modules}):
        yield PythonPathEnv(sys.path, sys.modules)



import platform
from enum import IntEnum, unique
from functools import lru_cache

from ...model import int_enum_loads

__all__ = [
    'is_cpython',
    'is_ironpython',
    'is_jython',
    'is_pypy',
]


@int_enum_loads(name_preprocess=str.upper)
@unique
class PythonImplementation(IntEnum):
    CPYTHON = 1
    IRONPYTHON = 2
    JYTHON = 3
    PYPY = 4


@lru_cache()
def _get_python_implementation() -> PythonImplementation:
    return PythonImplementation.loads(platform.python_implementation())


def is_cpython() -> bool:
    """
    Overview:
        Return ``True`` is current python is CPython, otherwise return ``False``.

    :return: Current python is CPython or not.
    """
    return _get_python_implementation() == PythonImplementation.CPYTHON


def is_ironpython() -> bool:
    """
    Overview:
        Return ``True`` is current python is IronPython, otherwise return ``False``.

    :return: Current python is IronPython or not.
    """
    return _get_python_implementation() == PythonImplementation.IRONPYTHON


def is_jython() -> bool:
    """
    Overview:
        Return ``True`` is current python is Jython, otherwise return ``False``.

    :return: Current python is Jython or not.
    """
    return _get_python_implementation() == PythonImplementation.JYTHON


def is_pypy() -> bool:
    """
    Overview:
        Return ``True`` is current python is PyPy, otherwise return ``False``.

    :return: Current python is PyPy or not.
    """
    return _get_python_implementation() == PythonImplementation.PYPY



"""
Overview:
    Useful functions to deal with exception or backtrace objects.
"""
import io
import traceback

__all__ = [
    'print_traceback',
    'str_traceback',
]


def print_traceback(err: BaseException, file=None):
    """
    Overview:
        Print full backtrace for exception object.
    Arguments:
        - err (:obj:`BaseException`): Exception object.
        - file: Output file, default is ``None`` which means stdout.

    Examples::
        See :func:`str_traceback` for actual print content.
    """
    traceback.print_exception(type(err), err, err.__traceback__, file=file)


def str_traceback(err: BaseException) -> str:
    """
    Overview:
        Get full backtrace for exception object.
    Arguments:
        - err (:obj:`BaseException`): Exception object.
    Returns:
        - backtrace (:obj:`str`): Full string backtrace.

    Examples:
        >>> try:
        >>>     raise RuntimeError('runtime error')
        >>> except RuntimeError as e:
        >>>     s = str_traceback(e)

        >>> s
        Traceback (most recent call last):
          File "<stdin>", line 2, in <module>
            raise RuntimeError('runtime error')
        RuntimeError: runtime error
    """
    with io.StringIO() as fs:
        print_traceback(err, file=fs)
        return fs.getvalue()
